"""Provides coordinate aware filters and smoothing."""

import copy

import numpy as np
from scipy import ndimage

import xarray as xr
from arpes.provenance import provenance
from arpes.utilities import lift_spectrum

__all__ = (
    "gaussian_filter_arr",
    "gaussian_filter",
    "boxcar_filter_arr",
    "boxcar_filter",
)


@lift_spectrum
def gaussian_filter_arr(
    spectrum: xr.DataArray, sigma=None, n=1, default_size=1
) -> xr.DataArray:
    """Coordinate aware `scipy.ndimage.filters.gaussian_filter`.

    Args:
        data: Data to smooth.
        sigma: Kernel sigma, specified in terms of axis units. An axis that is not specified
          will have a kernel width of `default_size` in index units.
        n: Repeats n times.
        default_size: Changes the default kernel width for axes not specified in `sigma`. Changing this
          parameter and leaving `sigma` as None allows you to smooth with an even-width
          kernel in index-coordinates.

    Returns:
        Smoothed data.
    """

    sigma = {} if sigma is None else sigma
    sigma = {
        k: int(v / (spectrum.coords[k][1] - spectrum.coords[k][0]))
        for k, v in sigma.items()
    }
    for dim in spectrum.dims:
        if dim not in sigma:
            sigma[dim] = default_size
    sigma = tuple(sigma[k] for k in spectrum.dims)

    values = spectrum.values
    for _ in range(n):
        values = ndimage.filters.gaussian_filter(values, sigma)

    filtered_spectrum = xr.DataArray(values, spectrum.coords, spectrum.dims)
    return filtered_spectrum


def gaussian_filter(sigma=None, n=1):
    """A partial application of `gaussian_filter_arr` that can be passed to derivative analysis functions.

    Args:
        sigma
        n

    Returns:
        A function which applies the Gaussian filter.
    """

    def f(arr):
        return gaussian_filter_arr(arr, sigma, n)

    return f


def boxcar_filter(size=None, n=1):
    """A partial application of `boxcar_filter_arr` that can be passed to derivative analysis functions.

    Args:
        size
        n

    Returns:
        A function which applies the boxcar.
    """

    def f(arr):
        return boxcar_filter_arr(arr, size, n)

    return f


def boxcar_filter_arr(
    arr: xr.DataArray, size=None, n=1, default_size=1, skip_nan=True
) -> xr.DataArray:
    """Coordinate aware `scipy.ndimage.filters.boxcar_filter`.

    Args:
        arr
        size: Kernel size, specified in terms of axis units. An axis
            that is not specified will have a kernel width of
            `default_size` in index units.
        n: Repeats n times.
        default_size: Changes the default kernel width for axes not
            specified in `sigma`. Changing this parameter and leaving
            `sigma` as None allows you to smooth with an even-width
            kernel in index-coordinates.
        skip_nan: By default, masks parts of the data which are NaN to
            prevent poor filter results.

    Returns:
        smoothed data.
    """
    if size is None:
        size = {}

    size = {k: int(v / (arr.coords[k][1] - arr.coords[k][0])) for k, v in size.items()}
    for dim in arr.dims:
        if dim not in size:
            size[dim] = default_size

    size = tuple(size[k] for k in arr.dims)

    if skip_nan:
        nan_mask = np.copy(arr.values) * 0 + 1
        nan_mask[arr.values != arr.values] = 0
        filtered_mask = ndimage.filters.uniform_filter(nan_mask, size)

        values = np.copy(arr.values)
        values[values != values] = 0

        for _ in range(n):
            values = ndimage.filters.uniform_filter(values, size) / filtered_mask
            values[nan_mask == 0] = 0
    else:
        for i in range(n):
            values = ndimage.filters.uniform_filter(values, size)

    filtered_arr = xr.DataArray(
        values, arr.coords, arr.dims, attrs=copy.deepcopy(arr.attrs)
    )

    if "id" in arr.attrs:
        del filtered_arr.attrs["id"]

        provenance(
            filtered_arr,
            arr,
            {
                "what": "Boxcar filtered data",
                "by": "boxcar_filter_arr",
                "size": size,
                "skip_nan": skip_nan,
            },
        )

    return filtered_arr



"""Wraps standard lmfit models."""
import lmfit as lf
import numpy as np

from lmfit.models import guess_from_peak, update_param_vals

from .x_model_mixin import XModelMixin

__all__ = [
    "VoigtModel",
    "GaussianModel",
    "ConstantModel",
    "LorentzianModel",
    "SkewedVoigtModel",
    "SplitLorentzianModel",
    "LinearModel",
    "LogisticModel",
    "StepModel",
    "SineModel",
]


class SineModel(XModelMixin, lf.models.SineModel):
    """Wraps `lf.models.SineModel`."""

    pass


class VoigtModel(XModelMixin, lf.models.VoigtModel):
    """Wraps `lf.models.VoigtModel`."""

    pass


class GaussianModel(XModelMixin, lf.models.GaussianModel):
    """Wraps `lf.models.GaussianModel`."""

    pass


class ConstantModel(XModelMixin, lf.models.ConstantModel):
    """Wraps `lf.models.ConstantModel`."""

    pass


class LorentzianModel(XModelMixin, lf.models.LorentzianModel):
    """Wraps `lf.models.LorentzianModel`."""

    pass


class SkewedVoigtModel(XModelMixin, lf.models.SkewedVoigtModel):
    """Wraps `lf.models.SkewedVoigtModel`."""

    pass


class SkewedGaussianModel(XModelMixin, lf.models.SkewedGaussianModel):
    """Wraps `lf.models.SkewedGaussianModel`."""

    pass


class SplitLorentzianModel(XModelMixin, lf.models.SplitLorentzianModel):
    """Wraps `lf.models.SplitLorentzianModel`."""

    def guess(self, data, x=None, **kwargs):
        """Estimate initial model parameter values from data."""
        pars = self.make_params()
        pars = guess_from_peak(self, data, x, negative=False, ampscale=1.25)
        sigma = pars["%ssigma" % self.prefix]
        pars["%ssigma_r" % self.prefix].set(value=sigma.value, min=sigma.min, max=sigma.max)

        return update_param_vals(pars, self.prefix, **kwargs)


class LinearModel(XModelMixin, lf.models.LinearModel):
    """A linear regression model."""

    def guess(self, data, x=None, **kwargs):
        """Use np.polyfit to get good initial parameters."""
        sval, oval = 0.0, 0.0
        if x is not None:
            sval, oval = np.polyfit(x, data, 1)
        pars = self.make_params(intercept=oval, slope=sval)
        return update_param_vals(pars, self.prefix, **kwargs)


class LogisticModel(XModelMixin, lf.models.StepModel):
    """A logistic regression model."""

    def __init__(self, independent_vars=["x"], prefix="", missing="raise", name=None, **kwargs):
        """Set standard parameters and delegate to lmfit."""
        kwargs.update(
            {
                "prefix": prefix,
                "missing": missing,
                "independent_vars": independent_vars,
                "form": "logistic",
            }
        )
        super().__init__(**kwargs)


class StepModel(XModelMixin, lf.models.StepModel):
    """Wraps `lf.models.StepModel`."""

    pass



"""Utilities related to computing Brillouin zones, masks, and selecting data.

TODO: Standardize this module around support for some other library that has proper
Brillouin zone plotting, like in ASE.

This module also includes tools for masking regions of data against
Brillouin zones.
"""

import itertools
import re
from collections import Counter, namedtuple

import matplotlib.path
import numpy as np

__all__ = (
    "bz_symmetry",
    "bz_cutter",
    "reduced_bz_selection",
    "reduced_bz_axes",
    "reduced_bz_mask",
    "reduced_bz_poly",
    "reduced_bz_axis_to",
    "reduced_bz_E_mask",
    "axis_along",
    "hex_cell",
    "hex_cell_2d",
    "orthorhombic_cell",
    "process_kpath",
)


_SYMMETRY_TYPES = {
    ("G", "X", "Y"): "rect",
    ("G", "X"): "square",
    ("G", "X", "BX"): "hex",
}

_POINT_NAMES_FOR_SYMMETRY = {
    "rect": {"G", "X", "Y"},
    "square": {"G", "X"},
    "hex": {"G", "X", "BX"},
}

SpecialPoint = namedtuple("SpecialPoint", ("name", "negate", "bz_coord"))


def as_3d(points):
    """Takes a 2D points list and zero pads to convert to a 3D representation."""
    return np.concatenate([points, points[:, 0][:, None] * 0], axis=1)


def as_2d(points):
    """Takes a 3D points and converts to a 2D representation by dropping the z coordinates."""
    return points[:, :2]


def parse_single_path(path):
    """Converts a path given by high symmetry point names to numerical coordinate arrays."""
    # first tokenize
    tokens = [name for name in re.split(r"([A-Z][a-z0-9]*(?:\([0-9,\s]+\))?)", path) if name]

    # normalize Gamma to G
    tokens = [token.replace("Gamma", "G") for token in tokens]

    # convert to standard format
    points = []
    for token in tokens:
        name, rest = token[0], token[1:]
        negate = False
        if rest and rest[0] == "n":
            negate = True
            rest = rest[1:]

        bz_coords = (
            0,
            0,
            0,
        )
        if rest:
            rest = "".join(c for c in rest if c not in "( \t\n\r)")
            bz_coords = tuple([int(c) for c in rest.split(",")])

        if len(bz_coords) == 2:
            bz_coords = tuple(list(bz_coords) + [0])
        points.append(SpecialPoint(name=name, negate=negate, bz_coord=bz_coords))

    return points


def parse_path(paths):
    """Converts paths to arrays with the coordinate locations for those paths."""
    if isinstance(paths, str):

        # some manual string work in order to make sure we do not split on commas inside BZ indices
        idxs = []
        for i, p in enumerate(paths):
            if p == ",":
                c = Counter(paths[:i])
                if c["("] - c[")"] == 0:
                    idxs.append(i)

        paths = list(paths)
        for idx in idxs:
            paths[idx] = ":"

        paths = "".join(paths)
        paths = paths.split(":")

    return [parse_single_path(p) for p in paths]


def special_point_to_vector(special_point, icell, special_points):
    """Converts a single special point to its coordinate vector."""
    base = np.dot(icell.T, special_points[special_point.name])

    if special_point.negate:
        base = -np.array(base)

    coord = np.array(special_point.bz_coord)
    return base + coord.dot(icell)


def process_kpath(paths, cell, special_points=None):
    """Converts paths consistign of point definitions to raw coordinates."""
    if len(cell) == 2:
        cell = [c + [0] for c in cell] + [0, 0, 0]

    icell = np.linalg.inv(cell).T

    if special_points is None:
        from ase.dft.kpoints import get_special_points  # pylint: disable=import-error

        special_points = get_special_points(cell)

    points = [
        [special_point_to_vector(elem, icell, special_points) for elem in p]
        for p in parse_path(paths)
    ]

    return points


# Some common Brillouin zone formats
def orthorhombic_cell(a=1, b=1, c=1):
    """Lattice constants for an orthorhombic unit cell."""
    return [[a, 0, 0], [0, b, 0], [0, 0, c]]


def hex_cell(a=1, c=1):
    """Calculates lattice vectors for a triangular lattice with lattice constants `a` and `c`."""
    return [[a, 0, 0], [-0.5 * a, 3 ** 0.5 / 2 * a, 0], [0, 0, c]]


def hex_cell_2d(a=1):
    """Calculates lattice vectors for a triangular lattice with lattice constant `a`."""
    return [[a, 0], [-0.5 * a, 3 ** 0.5 / 2 * a]]


def flat_bz_indices_list(bz_indices_list=None):
    """Calculate a flat representation of a repeated Brillouin zone specification.

    This is useful for plotting extra Brillouin zones or generating high symmetry points,
    lines, and planes.

    If None is provided, the first BZ is assumed.

    ```
    None -> [(0,0)]
    ```

    If an explicit zone is provided or a list of zones is provided, these are
    returned

    ```
    [(0,1,0), (-1, -1, 2)] -> [(0,1,0), (-1, -1, 2)]
    ```

    Additionally, tuples are unpacked into ranges

    ```
    [((-2, 1), 1)] -> [(-2, 1), (-1, 1), (0, 1)]
    ```
    """
    if bz_indices_list is None:
        bz_indices_list = [(0, 0)]

    try:
        l = len(bz_indices_list[0])
        if l not in {2, 3}:
            raise ValueError()
    except (ValueError, TypeError):
        bz_indices_list = [bz_indices_list]

    indices = []
    if len(bz_indices_list[0]) == 2:
        for bz_x, bz_y in bz_indices_list:
            rx = range(bz_x, bz_x + 1) if isinstance(bz_x, int) else range(*bz_x)
            ry = range(bz_y, bz_y + 1) if isinstance(bz_y, int) else range(*bz_y)
            for x, y in itertools.product(rx, ry):
                indices.append(
                    (
                        x,
                        y,
                    )
                )
    else:
        for bz_x, bz_y, bz_z in bz_indices_list:
            rx = range(bz_x, bz_x + 1) if isinstance(bz_x, int) else range(*bz_x)
            ry = range(bz_y, bz_y + 1) if isinstance(bz_y, int) else range(*bz_y)
            rz = range(bz_z, bz_z + 1) if isinstance(bz_z, int) else range(*bz_z)
            for x, y, z in itertools.product(rx, ry, rz):
                indices.append((x, y, z))

    return indices


def generate_2d_equivalent_points(points, icell, bz_indices_list=None):
    """Generates the equivalent points in higher order Brillouin zones."""
    points_list = []
    for x, y in flat_bz_indices_list(bz_indices_list):
        points_list.append(
            points[:, :2]
            + x
            * icell[0][
                None,
                :2,
            ]
            + y
            * icell[1][
                None,
                :2,
            ]
        )

    return np.unique(np.concatenate(points_list), axis=0)


def build_2dbz_poly(vertices=None, icell=None, cell=None):
    """Converts brillouin zone or equivalent information to a polygon mask.

    This mask can be used to mask away data outside the zone boundary.
    """
    from arpes.analysis.mask import raw_poly_to_mask
    from ase.dft.bz import bz_vertices  # pylint: disable=import-error

    assert cell is not None or vertices is not None or icell is not None

    if vertices is None:
        if icell is None:
            icell = np.linalg.inv(cell).T

        vertices = bz_vertices(icell)

    points, _ = vertices[0]  # points, normal
    points_2d = [p[:2] for p in points]

    return raw_poly_to_mask(points_2d)


def bz_symmetry(flat_symmetry_points):
    """Determines symmetry from a list of the symmetry points."""
    if isinstance(flat_symmetry_points, dict):
        flat_symmetry_points = flat_symmetry_points.items()

    largest_identified = 0
    symmetry = None

    point_names = set(k for k, _ in flat_symmetry_points)

    for points, sym in _SYMMETRY_TYPES.items():
        if all(p in point_names for p in points):
            if len(points) > largest_identified:
                symmetry = sym
                largest_identified = len(points)

    return symmetry


def reduced_bz_axis_to(data, S, include_E=False):
    """Calculates a displacement vector to a modded high symmetry point."""
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}

    coords_by_point = {
        k: np.array([v.get(d, 0) for d in data.dims if d in v.keys() or include_E and d == "eV"])
        for k, v in points.items()
    }
    if symmetry == "rect":
        if S == "X":
            return coords_by_point["X"] - coords_by_point["G"]
        return coords_by_point["Y"] - coords_by_point["G"]
    elif symmetry == "square":
        raise NotImplementedError
        return coords_by_point["X"] - coords_by_point["G"]
    elif symmetry == "hex":
        if S == "X":
            return coords_by_point["X"] - coords_by_point["G"]
        return coords_by_point["BX"] - coords_by_point["G"]
    else:
        raise NotImplementedError


def reduced_bz_axes(data):
    """Calculates displacement vectors to high symmetry points in the first Brillouin zone."""
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}

    coords_by_point = {
        k: np.array([v[d] for d in data.dims if d in v.keys()]) for k, v in points.items()
    }
    if symmetry == "rect":
        dx = coords_by_point["X"] - coords_by_point["G"]
        dy = coords_by_point["Y"] - coords_by_point["G"]
    elif symmetry == "square":
        raise NotImplementedError
        dx = coords_by_point["X"] - coords_by_point["G"]
        dy = coords_by_point["X"] - coords_by_point["G"]
    elif symmetry == "hex":
        dx = coords_by_point["X"] - coords_by_point["G"]
        dy = coords_by_point["BX"] - coords_by_point["G"]
    else:
        raise NotImplementedError

    return dx, dy


def axis_along(data, S):
    """Determines which axis lies principally along the direction G->S."""
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}

    coords_by_point = {
        k: np.array([v[d] for d in data.dims if d in v.keys()]) for k, v in points.items()
    }

    dS = coords_by_point[S] - coords_by_point["G"]

    max = -np.inf
    max_dim = None
    for dD, d in zip(dS, [d for d in data.dims if d != "eV"]):
        if np.abs(dD) > max:
            max = np.abs(dD)
            max_dim = d

    return max_dim


def reduced_bz_poly(data, scale_zone=False):
    """Returns a polynomial representing the reduce first Brillouin zone."""
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]
    dx, dy = reduced_bz_axes(data)
    if scale_zone:
        # should be good enough, reevaluate later
        dx = 3 * dx
        dy = 3 * dy

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}
    coords_by_point = {
        k: np.array([v.get(d, 0) for d in data.dims if d in v.keys()]) for k, v in points.items()
    }

    if symmetry == "hex":
        return np.array(
            [
                coords_by_point["G"],
                coords_by_point["G"] + dx,
                coords_by_point["G"] + dy,
            ]
        )

    return np.array(
        [
            coords_by_point["G"],
            coords_by_point["G"] + dx,
            coords_by_point["G"] + dx + dy,
            coords_by_point["G"] + dy,
        ]
    )


def reduced_bz_E_mask(data, S, e_cut, scale_zone=False):
    """Calculates a mask for data which contains points below an energy cutoff."""
    symmetry_points, _ = data.S.symmetry_points()
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]
    # bz_dims = tuple(d for d in data.dims if d in list(symmetry_points.values())[0][0].keys())

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}
    coords_by_point = {
        k: np.array([v.get(d, 0) for d in data.dims if d in v.keys() or d == "eV"])
        for k, v in points.items()
    }

    dx_to = reduced_bz_axis_to(data, S, include_E=True)
    if scale_zone:
        dx_to = dx_to * 3
    dE = np.array([0 if d != "eV" else e_cut for d in data.dims])

    poly_points = np.array(
        [
            coords_by_point["G"],
            coords_by_point["G"] + dx_to,
            coords_by_point["G"] + dx_to + dE,
            coords_by_point["G"] + dE,
        ]
    )

    skip_col = None
    for i in range(poly_points.shape[1]):
        if np.all(poly_points[:, i] == poly_points[0, i]):
            skip_col = i

    assert skip_col is not None
    selector_val = poly_points[0, skip_col]
    poly_points = np.concatenate(
        (poly_points[:, 0:skip_col], poly_points[:, skip_col + 1 :]), axis=1
    )

    selector = dict()
    selector[data.dims[skip_col]] = selector_val
    sdata = data.sel(**selector, method="nearest")

    path = matplotlib.path.Path(poly_points)
    grid = np.array(
        [a.ravel() for a in np.meshgrid(*[data.coords[d] for d in sdata.dims], indexing="ij")]
    ).T
    mask = path.contains_points(grid)
    mask = np.reshape(mask, sdata.data.shape)
    return mask


def reduced_bz_mask(data, **kwargs):
    """Calculates a mask for the first Brillouin zone of a piece of data."""
    symmetry_points, _ = data.S.symmetry_points()
    bz_dims = tuple(d for d in data.dims if d in list(symmetry_points.values())[0][0].keys())

    poly_points = reduced_bz_poly(data, **kwargs)
    extra_dims_shape = tuple(len(data.coords[d]) for d in data.dims if d in bz_dims)

    path = matplotlib.path.Path(poly_points)
    grid = np.array(
        [a.ravel() for a in np.meshgrid(*[data.coords[d] for d in bz_dims], indexing="ij")]
    ).T
    mask = path.contains_points(grid)
    mask = np.reshape(mask, extra_dims_shape)

    return mask


def reduced_bz_selection(data):
    """Sets data outisde the Brillouin zone mask for a piece of data to be nan."""
    mask = reduced_bz_mask(data)

    data = data.copy()
    data.data[np.logical_not(mask)] = np.nan

    return data


def bz_cutter(symmetry_points, reduced=True):
    """Cuts data so that it areas outside the Brillouin zone are masked away.

    TODO: UNFINISHED.
    """

    def build_bz_mask(data):
        pass

    def cutter(data, cut_value=np.nan):
        mask = build_bz_mask(data)

        out = data.copy()
        out.data[mask] = cut_value

        return out

    return cutter



"""Provides a Qt based implementation of a curve fit inspection tool."""

from arpes.plotting.qt_tool.BinningInfoWidget import BinningInfoWidget
from arpes.utilities.qt.utils import PlotOrientation, ReactivePlotRecord
from PyQt5 import QtGui, QtCore, QtWidgets
import pyqtgraph as pg
import numpy as np
import dill
from dataclasses import dataclass
import enum
import xarray as xr
import weakref
import warnings
from typing import List, Optional, Union

from arpes.utilities.qt.data_array_image_view import DataArrayPlot

from arpes.fits.utilities import result_to_hints
from arpes.utilities.qt.ui import (
    KeyBinding,
    CursorRegion,
    button,
    horizontal,
    label,
    tabs,
)
from arpes.utilities.qt import (
    qt_info,
    DataArrayImageView,
    BasicHelpDialog,
    SimpleWindow,
    SimpleApp,
    run_tool_in_daemon_process,
)

from .fit_inspection_plot import FitInspectionPlot

__all__ = (
    "FitTool",
    "fit_tool",
)

qt_info.setup_pyqtgraph()


class DataKey(str, enum.Enum):
    Data = "data"
    Residual = "residual"
    NormalizedResidual = "norm_residual"


class FitToolWindow(SimpleWindow):
    """The application window for `FitTool`."""

    HELP_DIALOG_CLS = BasicHelpDialog

    def compile_key_bindings(self):
        return super().compile_key_bindings() + [
            KeyBinding(
                "Scroll Cursor",
                [
                    QtCore.Qt.Key_Left,
                    QtCore.Qt.Key_Right,
                    QtCore.Qt.Key_Up,
                    QtCore.Qt.Key_Down,
                ],
                self.scroll,
            ),
            KeyBinding(
                "Reset Intensity",
                [QtCore.Qt.Key_I],
                self.reset_intensity,
            ),
            KeyBinding(
                "Center Cursor",
                [QtCore.Qt.Key_C],
                self.center_cursor,
            ),
            KeyBinding(
                "Transpose - Roll Axis",
                [QtCore.Qt.Key_T],
                self.transpose_roll,
            ),
            KeyBinding(
                "Transpose - Swap Front Axes",
                [QtCore.Qt.Key_Y],
                self.transpose_swap,
            ),
        ]

    def center_cursor(self, event):
        self.app().center_cursor()

    def transpose_roll(self, event):
        self.app().transpose_to_front(-1)

    def transpose_swap(self, event):
        self.app().transpose_to_front(1)

    @staticmethod
    def _update_scroll_delta(delta, event: QtGui.QKeyEvent):
        if event.nativeModifiers() & 1:  # shift key
            delta = (delta[0], delta[1] * 5)

        if event.nativeModifiers() & 2:  # shift key
            delta = (delta[0], delta[1] * 20)

        return delta

    def reset_intensity(self, event: QtGui.QKeyEvent):
        self.app().reset_intensity()

    def scroll_z(self, event: QtGui.QKeyEvent):
        key_map = {
            QtCore.Qt.Key_N: (2, -1),
            QtCore.Qt.Key_M: (2, 1),
        }

        delta = self._update_scroll_delta(key_map.get(event.key()), event)

        if delta is not None and self.app() is not None:
            self.app().scroll(delta)

    def scroll(self, event: QtGui.QKeyEvent):
        key_map = {
            QtCore.Qt.Key_Left: (0, -1),
            QtCore.Qt.Key_Right: (0, 1),
            QtCore.Qt.Key_Down: (1, -1),
            QtCore.Qt.Key_Up: (1, 1),
        }

        delta = self._update_scroll_delta(key_map.get(event.key()), event)

        if delta is not None and self.app() is not None:
            self.app().scroll(delta)


@dataclass
class FitTool(SimpleApp):
    """FitTool is an implementation of a curve fit browser for PyARPES."""

    data_key: DataKey = DataKey.Data
    dataset: Optional[xr.Dataset] = None

    TITLE = "Fit Tool"
    WINDOW_CLS = FitToolWindow
    WINDOW_SIZE = (8, 5)

    def __init__(self):
        """Initialize attributes to safe empty values."""
        super().__init__()

        self.content_layout = None
        self.main_layout = None

    @property
    def data(self) -> xr.DataArray:
        """Extract the array-like values according to what content we are rendering."""
        return self.dataset[self.data_key.value]

    @data.setter
    def data(self, new_data) -> None:
        raise TypeError("On fit_tool, the data is computed from the original dataset.")

    def center_cursor(self):
        """Scrolls so that the cursors are in the center of the data volume."""
        new_cursor = [len(self.data.coords[d]) / 2 for d in self.data.dims]
        self.update_cursor_position(new_cursor)

        for i, cursors in self.registered_cursors.items():
            for cursor in cursors:
                cursor.set_location(new_cursor[i])

    def scroll(self, delta):
        """Scroll the axis delta[0] by delta[1] pixels."""
        if delta[0] >= len(self.context["cursor"]):
            warnings.warn("Tried to scroll a non-existent dimension.")
            return

        cursor = list(self.context["cursor"])
        cursor[delta[0]] += delta[1]

        self.update_cursor_position(cursor)

        for i, cursors in self.registered_cursors.items():
            for c in cursors:
                c.set_location(cursor[i])

    def transpose(self, transpose_order: List[str]):
        """Transpose dimensions into the order specified by `transpose_order` and redraw."""
        reindex_order = [self.data.dims.index(t) for t in transpose_order]
        self.data = self.data.transpose(*transpose_order)

        new_cursor = [self.context["cursor"][i] for i in reindex_order]
        self.update_cursor_position(new_cursor, force=True)

        for i, cursors in self.registered_cursors.items():
            for cursor in cursors:
                cursor.set_location(new_cursor[i])

    def transpose_to_front(self, dim: Union[str, int]):
        """Transpose the dimension `dim` to the front so that it is in the main marginal."""
        if not isinstance(dim, str):
            dim = self.data.dims[dim]

        order = list(self.data.dims)
        order.remove(dim)
        order = [dim] + order
        self.transpose(order)

    def configure_image_widgets(self):
        """Configure array marginals for the input data.

        Depending on the array dimensionality, we need a different number and variety
        of marginals. This is as easy as specifying which marginals we select over and
        handling the rest dynamically.

        An additional complexity is that we also handle the cursor registration here.

        Unlike the simple data browser, we have a few additional complexities. For one,
        we need to generate a display of the fit which is currently under the cursor.

        For now, we are going to support 1D fits only. The remaining data will be either
        1, 2, or 3 dimensional, under current assumptions. Let's consider each case in
        turn.

        One Dimensional Marginals:
            In this case, the total dataset is 2D, we will put a single 1D cursor
            on the broadcast axis but we will display *all* of the data as a 2D image
            plot with a 1D marginal and the 1D fit display.

        Two Dimensional Marginals:
            Here, we will give a single 2D image plot with the perpendicular (fit)
            axis out of the image plane. A cursor will be registered to the 1D fit
            display so that the contents of the main display can be changed. This 1D
            fit cursor will have binning controls.

        Three Dimensional Marginals:
            The total dataset is 4D. We will use the standard 3D set of marginal planes,
            and 1D marginals like we do in the 3D qt_tool with a 3D cursor.

            The 1D marginal will have a cursor and binning controls on that cursor.
        """
        if len(self.data.dims) == 2:  # 1 broadcast dimension and one data dimension
            self.generate_marginal_for(
                (), 0, 0, "xy", cursors=True, layout=self.content_layout
            )
            self.generate_fit_marginal_for(
                (0, 1),
                0,
                1,
                "fit",
                cursors=False,
                orientation=PlotOrientation.Vertical,
                layout=self.content_layout,
            )
            self.views["xy"].view.setYLink(self.views["fit"].inner_plot)

        if len(self.data.dims) == 3:
            self.generate_marginal_for(
                (2,), 1, 0, "xy", cursors=True, layout=self.content_layout
            )
            self.generate_fit_marginal_for(
                (0, 1, 2), 0, 0, "fit", cursors=True, layout=self.content_layout
            )

        if len(self.data.dims) == 4:
            # no idea if these marginal locations are correct, need to check that
            self.generate_marginal_for(
                (1, 3), 1, 0, "xz", cursors=True, layout=self.content_layout
            )
            self.generate_marginal_for(
                (2, 3), 0, 1, "xy", cursors=True, layout=self.content_layout
            )
            self.generate_marginal_for((0, 3), 1, 1, "yz", layout=self.content_layout)
            self.generate_fit_marginal_for(
                (0, 1, 2, 3), 0, 0, "fit", cursors=True, layout=self.content_layout
            )

    def generate_fit_marginal_for(
        self,
        dimensions,
        column,
        row,
        name="fit",
        orientation=PlotOrientation.Horizontal,
        cursors=False,
        layout=None,
    ):
        """Generates a marginal plot for a fit at a given set of coordinates.

        This does something very similar to `generate_marginal_for` except that it is
        specialized to showing a widget which embeds information about the current fit result.
        """
        if layout is None:
            layout = self._layout

        remaining_dims = [
            l for l in list(range(len(self.data.dims))) if l not in dimensions
        ]

        # for now, we only allow a single fit dimension
        widget = FitInspectionPlot(
            name=name, root=weakref.ref(self), orientation=orientation
        )
        self.views[name] = widget

        if orientation == PlotOrientation.Horizontal:
            widget.setMaximumHeight(qt_info.inches_to_px(3))
        else:
            widget.setMaximumWidth(qt_info.inches_to_px(3))

        if cursors:
            cursor = CursorRegion(
                orientation=(
                    CursorRegion.Vertical
                    if orientation == PlotOrientation.Vertical
                    else CursorRegion.Horizontal
                ),
                movable=True,
            )
            widget.addItem(cursor, ignoreBounds=False)
            self.connect_cursor(remaining_dims[-1], cursor)

        self.reactive_views.append(
            ReactivePlotRecord(dims=dimensions, view=widget, orientation=orientation)
        )
        layout.addWidget(widget, column, row)
        return widget

    def connect_cursor(self, dimension, the_line):
        """Connect a cursor to a line control.

        without weak references we get a circular dependency here
        because `the_line` is owned by a child of `self` but we are
        providing self to a closure which is retained by `the_line`.
        """
        self.registered_cursors[dimension].append(the_line)
        owner = weakref.ref(self)

        def connected_cursor(line: CursorRegion):
            new_cursor = list(owner().context["cursor"])
            new_cursor[dimension] = line.getRegion()[0]
            owner().update_cursor_position(new_cursor)

        the_line.sigRegionChanged.connect(connected_cursor)

    def update_cursor_position(self, new_cursor, force=False, keep_levels=True):
        """Sets the current cursor position.

        Because setting the cursor position changes the marginal data, this is also
        where redrawing originates.

        The way we do this is basically to step through views, recompute the slice for that view
        and set the image/array on the slice.
        """
        old_cursor = list(self.context["cursor"])
        self.context["cursor"] = new_cursor

        def index_to_value(value, i):
            d = self.data.dims[i]
            c = self.data.coords[d].values
            return c[0] + value * (c[1] - c[0])

        self.context["value_cursor"] = [
            index_to_value(v, i) for i, v in enumerate(new_cursor)
        ]

        changed_dimensions = [
            i for i, (x, y) in enumerate(zip(old_cursor, new_cursor)) if x != y
        ]

        cursor_text = ",".join(
            "{}: {:.4g}".format(x, y)
            for x, y in zip(self.data.dims, self.context["value_cursor"])
        )
        self.window.statusBar().showMessage("({})".format(cursor_text))

        # update data
        def safe_slice(vlow, vhigh, axis=0):
            vlow, vhigh = int(min(vlow, vhigh)), int(max(vlow, vhigh))
            rng = len(self.data.coords[self.data.dims[axis]])
            vlow, vhigh = np.clip(vlow, 0, rng), np.clip(vhigh, 0, rng)

            if vlow == vhigh:
                vhigh = vlow + 1

            vlow, vhigh = np.clip(vlow, 0, rng), np.clip(vhigh, 0, rng)

            if vlow == vhigh:
                vlow = vhigh - 1

            return slice(vlow, vhigh)

        for reactive in self.reactive_views:
            if set(reactive.dims).intersection(set(changed_dimensions)) or force:
                try:
                    select_coord = dict(
                        zip(
                            [self.data.dims[i] for i in reactive.dims],
                            [
                                safe_slice(
                                    int(new_cursor[i]), int(new_cursor[i] + 1), i
                                )
                                for i in reactive.dims
                            ],
                        )
                    )
                    if isinstance(reactive.view, DataArrayImageView):
                        image_data = self.data.isel(**select_coord)
                        if select_coord:
                            image_data = image_data.mean(list(select_coord.keys()))
                        reactive.view.setImage(image_data, keep_levels=keep_levels)
                    elif isinstance(reactive.view, FitInspectionPlot):
                        results_coord = {
                            k: v
                            for k, v in select_coord.items()
                            if k in self.dataset.results.dims
                        }
                        result = self.dataset.results.isel(**results_coord)
                        result = result.item()
                        reactive.view.set_model_result(result)

                    elif isinstance(reactive.view, pg.PlotWidget):
                        for_plot = self.data.isel(**select_coord)
                        if select_coord:
                            for_plot = for_plot.mean(list(select_coord.keys()))

                        cursors = [
                            l
                            for l in reactive.view.getPlotItem().items
                            if isinstance(l, CursorRegion)
                        ]
                        reactive.view.clear()
                        for c in cursors:
                            reactive.view.addItem(c)

                        if isinstance(reactive.view, DataArrayPlot):
                            reactive.view.plot(for_plot)
                            continue

                        if reactive.orientation == PlotOrientation.Horizontal:
                            reactive.view.plot(for_plot.values)
                        else:
                            reactive.view.plot(
                                for_plot.values, range(len(for_plot.values))
                            )
                except IndexError:
                    pass

    def construct_binning_tab(self):
        """Gives tab controls for the axis along the fit only."""
        inner_items = [
            BinningInfoWidget(axis_name=len(self.data.dims) - 1, root=weakref.ref(self))
        ]
        return horizontal(label("Options"), *inner_items), inner_items

    def copy_parameter_hint(self, *_) -> None:
        """Converts parameters for the current model being displayed and copies to clipboard."""
        result = self.views["fit"].result
        hint = result_to_hints(result)
        self.copy_to_clipboard(hint)

    def construct_info_tab(self):
        """Provides some utility functionality to make curve fitting easier."""
        copy_button = button("Copy parameters as hint")
        copy_button.setMaximumWidth(qt_info.inches_to_px(1.5))
        copy_button.subject.subscribe(self.copy_parameter_hint)
        inner_items = [copy_button]
        return horizontal(*inner_items), inner_items

    def add_contextual_widgets(self):
        """Adds the widgets for the contextual controls at the bottom."""
        self.main_layout.addLayout(self.content_layout, 0, 0)
        info_tab, self.info_tab_widgets = self.construct_info_tab()
        binning_tab, self.binning_tab_widgets = self.construct_binning_tab()

        self.tabs = tabs(["Info", info_tab], ["Binning", binning_tab])
        self.tabs.setFixedHeight(qt_info.inches_to_px(1))

        self.main_layout.addLayout(self.content_layout, 0, 0)
        self.main_layout.addWidget(self.tabs, 1, 0)

    def layout(self):
        """Initialize the layout components."""
        self.main_layout = QtWidgets.QGridLayout()
        self.content_layout = QtWidgets.QGridLayout()
        return self.main_layout

    def before_show(self):
        """Lifecycle hook for configuration before app show."""
        self.configure_image_widgets()
        self.add_contextual_widgets()
        import matplotlib.cm

        self.set_colormap(matplotlib.cm.viridis)

    def after_show(self):
        """Initialize application state after app show.

        To do this, we need to set the initial cursor location, and call update
        which forces a rerender.
        """
        # basic state initialization
        self.context.update(
            {
                "cursor": [self.data.coords[d].mean().item() for d in self.data.dims],
            }
        )

        # Display the data
        self.update_cursor_position(
            self.context["cursor"], force=True, keep_levels=False
        )
        self.center_cursor()

    def reset_intensity(self):
        """Autoscales intensity in each marginal plot."""
        self.update_cursor_position(
            self.context["cursor"], force=True, keep_levels=False
        )

    def set_data(self, data: xr.Dataset):
        """Sets the current data to a new value and resets UI state."""
        self.dataset = data
        self.data_key = DataKey.Data

        # For now, we only support 1D fit results
        fit_dims = self.dataset.F.fit_dimensions
        assert len(fit_dims) == 1
        self.dataset = self.dataset.S.transpose_to_back(*fit_dims)


def _fit_tool(data: xr.Dataset) -> None:
    """Starts the fitting inspection tool using an input fit result Dataset."""
    try:
        data = dill.loads(data)
    except TypeError:
        pass

    # some sanity checks that we were actually passed a collection of fit results
    assert isinstance(data, xr.Dataset)
    assert "results" in data.data_vars

    tool = FitTool()
    tool.set_data(data)
    tool.start()


fit_tool = run_tool_in_daemon_process(_fit_tool)



"""Utilities used in broadcast fitting."""

import lmfit
import xarray as xr
import operator
import warnings
import functools
from string import ascii_lowercase


def unwrap_params(params, iter_coordinate):
    """Inspects array-like parameters and extracts the appropriate value to use for the current fit."""

    def transform_or_walk(v):
        if isinstance(v, dict):
            return unwrap_params(v, iter_coordinate)

        if isinstance(v, xr.DataArray):
            return v.sel(**iter_coordinate, method="nearest").item()

        return v

    return {k: transform_or_walk(v) for k, v in params.items()}


def apply_window(data: xr.DataArray, cut_coords: dict[str, float | slice], window):
    """Cuts data inside a specified window.

    Because we allow passing an array of windows, we need to first try to find
    the right one for the current fit before application.

    If there's no window, this acts as the identity function on the data.
    """
    cut_data = data.sel(**cut_coords)
    original_cut_data = cut_data

    if isinstance(window, xr.DataArray):
        window_item = window.sel(**cut_coords).item()
        if isinstance(window_item, slice):
            cut_data = cut_data.sel(**dict([[cut_data.dims[0], window_item]]))

    return cut_data, original_cut_data


def _parens_to_nested(items):
    """Turns a flat list with parentheses tokens into a nested list."""
    parens = [
        (
            token,
            idx,
        )
        for idx, token in enumerate(items)
        if isinstance(token, str) and token in "()"
    ]
    if parens:
        first_idx, last_idx = parens[0][1], parens[-1][1]
        if parens[0][0] != "(" or parens[-1][0] != ")":
            raise ValueError("Parentheses do not match!")

        return (
            items[0:first_idx]
            + [_parens_to_nested(items[first_idx + 1 : last_idx])]
            + items[last_idx + 1 :]
        )
    else:
        return items


def reduce_model_with_operators(model):
    """Combine models according to mathematical operators."""
    if isinstance(model, tuple):
        return model[0](prefix="{}_".format(model[1]), nan_policy="omit")

    if isinstance(model, list) and len(model) == 1:
        return reduce_model_with_operators(model[0])

    left, op, right = model[0], model[1], model[2:]
    left, right = reduce_model_with_operators(left), reduce_model_with_operators(right)

    if op == "+":
        return left + right
    elif op == "*":
        return left * right
    elif op == "-":
        return left - right
    elif op == "/":
        return left / right


def compile_model(model, params=None, prefixes=None):
    """Generates an lmfit model instance from specification.

    Takes a model sequence, i.e. a Model class, a list of such classes, or a list
    of such classes with operators and instantiates an appropriate model.
    """
    if params is None:
        params = {}

    prefix_compile = "{}"
    if prefixes is None:
        prefixes = ascii_lowercase
        prefix_compile = "{}_"

    try:
        if issubclass(model, lmfit.Model):
            return model()
    except TypeError:
        pass

    if isinstance(model, (list, tuple)) and all(
        [isinstance(token, type) for token in model]
    ):
        models = [
            m(prefix=prefix_compile.format(prefixes[i]), nan_policy="omit")
            for i, m in enumerate(model)
        ]
        if isinstance(params, (list, tuple)):
            for cs, m in zip(params, models):
                for name, params_for_name in cs.items():
                    m.set_param_hint(name, **params_for_name)

        built = functools.reduce(operator.add, models)
    else:
        warnings.warn("Beware of equal operator precedence.")
        prefix = iter(prefixes)
        model = [m if isinstance(m, str) else (m, next(prefix)) for m in model]
        built = reduce_model_with_operators(_parens_to_nested(model))

    return built



"""Some common spatial plotting routines. Useful for contextualizing nanoARPES data."""
from typing import Any, Dict, List, Optional
import itertools

import numpy as np
import xarray as xr

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib import cm
import matplotlib.patheffects as path_effects
import matplotlib.gridspec as gridspec

from arpes.io import load_data
from arpes.plotting.annotations import annotate_point
from arpes.plotting.utils import (
    path_for_plot,
    frame_with,
    remove_colorbars,
    fancy_labels,
    ddata_daxis_units,
)
from arpes.provenance import save_plot_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.xarray import unwrap_xarray_item


__all__ = ("reference_scan_spatial", "plot_spatial_reference")


@save_plot_provenance
def plot_spatial_reference(
    reference_map: DataType,
    data_list: List[DataType],
    offset_list: Optional[List[Dict[str, Any]]] = None,
    annotation_list: Optional[List[str]] = None,
    out: Optional[str] = None,
    plot_refs: bool = True,
):
    """Helpfully plots data against a reference scanning dataset.

    This is essential to understand
    where data was taken and can be used early in the analysis phase in order to highlight the
    location of your datasets against core levels, etc.

    Args:
        reference_map: A scanning photoemission like dataset
        data_list: A list of datasets you want to plot the relative locations of
        offset_list: Optionally, offsets given as coordinate dicts
        annotation_list: Optionally, text annotations for the data
        out: Where to save the figure if we are outputting to disk
        plot_refs: Whether to plot reference figures for each of the pieces of data in `data_list`
    """
    if offset_list is None:
        offset_list = [None] * len(data_list)

    if annotation_list is None:
        annotation_list = [str(i + 1) for i in range(len(data_list))]

    normalize_to_spectrum(reference_map)

    n_references = len(data_list)
    if n_references == 1 and plot_refs:
        fig, axes = plt.subplots(1, 2, figsize=(12, 5))
        ax = axes[0]
        ax_refs = [axes[1]]
    elif plot_refs:
        n_extra_axes = 1 + (n_references // 4)
        fig = plt.figure(figsize=(6 * (1 + n_extra_axes), 5))
        spec = gridspec.GridSpec(ncols=2 * (1 + n_extra_axes), nrows=2, figure=fig)
        ax = fig.add_subplot(spec[:2, :2])

        ax_refs = [
            fig.add_subplot(spec[i // (2 * n_extra_axes), 2 + i % (2 * n_extra_axes)])
            for i in range(n_references)
        ]
    else:
        ax_refs = []
        fig, ax = plt.subplots(1, 1, figsize=(6, 5))

    try:
        reference_map = reference_map.S.spectra[0]
    except Exception:
        pass

    reference_map = reference_map.S.mean_other(["x", "y", "z"])

    ref_dims = reference_map.dims[::-1]

    assert len(reference_map.dims) == 2
    reference_map.S.plot(ax=ax, cmap="Blues")

    cmap = cm.get_cmap("Reds")
    rendered_annotations = []
    for i, (data, offset, annotation) in enumerate(zip(data_list, offset_list, annotation_list)):
        if offset is None:
            try:
                offset = data.S.logical_offsets - reference_map.S.logical_offsets
            except ValueError:
                offset = {}

        coords = {c: unwrap_xarray_item(data.coords[c]) for c in ref_dims}
        n_array_coords = len(
            [cv for cv in coords.values() if isinstance(cv, (np.ndarray, xr.DataArray))]
        )
        color = cmap(0.4 + (0.5 * i / len(data_list)))
        x = coords[ref_dims[0]] + offset.get(ref_dims[0], 0)
        y = coords[ref_dims[1]] + offset.get(ref_dims[1], 0)
        ref_x, ref_y = x, y
        off_x, off_y = 0, 0
        scale = 0.03

        if n_array_coords == 0:
            off_y = 1
            ax.scatter([x], [y], s=60, color=color)
        if n_array_coords == 1:
            if isinstance(x, (np.ndarray, xr.DataArray)):
                y = [y] * len(x)
                ref_x = np.min(x)
                off_x = -1
            else:
                x = [x] * len(y)
                ref_y = np.max(y)
                off_y = 1

            ax.plot(x, y, color=color, linewidth=3)
        if n_array_coords == 2:
            off_y = 1
            min_x, max_x = np.min(x), np.max(x)
            min_y, max_y = np.min(y), np.max(y)
            ref_x, ref_y = min_x, max_y

            color = cmap(0.4 + (0.5 * i / len(data_list)), alpha=0.5)
            rect = patches.Rectangle((min_x, min_y), max_x - min_x, max_y - min_y, facecolor=color)
            color = cmap(0.4 + (0.5 * i / len(data_list)))

            ax.add_patch(rect)

        dp = ddata_daxis_units(ax)
        text_location = (
            np.asarray(
                [
                    ref_x,
                    ref_y,
                ]
            )
            + dp * scale * np.asarray([off_x, off_y])
        )
        text = ax.annotate(annotation, text_location, color="black", size=15)
        rendered_annotations.append(text)
        text.set_path_effects(
            [path_effects.Stroke(linewidth=2, foreground="white"), path_effects.Normal()]
        )
        if plot_refs:
            ax_ref = ax_refs[i]
            keep_preference = list(ref_dims) + [
                "eV",
                "temperature",
                "kz",
                "hv",
                "kp",
                "kx",
                "ky",
                "phi",
                "theta",
                "beta",
                "pixel",
            ]
            keep = [d for d in keep_preference if d in data.dims][:2]
            data.S.mean_other(keep).S.plot(ax=ax_ref)
            ax_ref.set_title(annotation)
            fancy_labels(ax_ref)
            frame_with(ax_ref, color=color, linewidth=3)

    ax.set_title("")
    remove_colorbars()
    fancy_labels(ax)
    plt.tight_layout()

    try:
        from adjustText import adjust_text

        adjust_text(
            rendered_annotations,
            ax=ax,
            avoid_points=False,
            avoid_objects=False,
            avoid_self=False,
            autoalign="xy",
        )
    except ImportError:
        pass

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, [ax] + ax_refs


@save_plot_provenance
def reference_scan_spatial(data, out=None, **kwargs):
    """Plots the spatial content of a dataset, useful as a quick reference."""
    data = normalize_to_spectrum(data)

    dims = [d for d in data.dims if d in {"cycle", "phi", "eV"}]

    summed_data = data.sum(dims, keep_attrs=True)

    fig, ax = plt.subplots(3, 2, figsize=(15, 15))
    flat_axes = list(itertools.chain(*ax))

    summed_data.plot(ax=flat_axes[0])
    flat_axes[0].set_title(r"Full \textbf{eV} range")

    dims_except_eV = [d for d in dims if d != "eV"]
    summed_data = data.sum(dims_except_eV)

    mul = 0.2
    rng = data.coords["eV"].max().item() - data.coords["eV"].min().item()
    offset = data.coords["eV"].max().item()
    if offset > 0:
        offset = 0

    if rng > 3:
        mul = rng / 5.0

    for i in range(5):
        low_e, high_e = -mul * (i + 1) + offset, -mul * i + offset
        title = r"\textbf{eV}" + ": {:.2g} to {:.2g}".format(low_e, high_e)
        summed_data.sel(eV=slice(low_e, high_e)).sum("eV").plot(ax=flat_axes[i + 1])
        flat_axes[i + 1].set_title(title)

    y_range = flat_axes[0].get_ylim()
    x_range = flat_axes[0].get_xlim()
    delta_one_percent = ((x_range[1] - x_range[0]) / 100, (y_range[1] - y_range[0]) / 100)

    smart_delta = (2 * delta_one_percent[0], -1.5 * delta_one_percent[0])

    referenced = data.S.referenced_scans

    # idea here is to collect points by those that are close together, then
    # only plot one annotation
    condensed = []
    cutoff = 3  # 3 percent
    for index, _ in referenced.iterrows():
        ff = load_data(index)

        x, y, _ = ff.S.sample_pos
        found = False
        for cx, cy, cl in condensed:
            if abs(cx - x) < cutoff * abs(delta_one_percent[0]) and abs(cy - y) < cutoff * abs(
                delta_one_percent[1]
            ):
                cl.append(index)
                found = True
                break

        if not found:
            condensed.append((x, y, [index]))

    for fax in flat_axes:
        for cx, cy, cl in condensed:
            annotate_point(
                fax,
                (
                    cx,
                    cy,
                ),
                ",".join([str(l) for l in cl]),
                delta=smart_delta,
                fontsize="large",
            )

    plt.tight_layout()

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax



"""Annotations onto plots for experimental conditions or locations."""
import numpy as np

from arpes.plotting.utils import name_for_dim, unit_for_dim
from arpes.utilities.conversion.forward import convert_coordinates_to_kspace_forward

__all__ = (
    "annotate_cuts",
    "annotate_point",
    "annotate_experimental_conditions",
)


def annotate_experimental_conditions(ax, data, desc, show=False, orientation="top", **kwargs):
    """Renders information about the experimental conditions onto a set of axes.

    Also adjust the axes limits and hides the axes.

    data should be the dataset described, and desc should be one of

    'temp',
    'photon',
    'photon polarization',
    'polarization',

    or a number to act as a spacer in units of the axis coordinates

    or a list of such items.
    """
    if isinstance(desc, (str, int, float)):
        desc = [desc]

    ax.grid(False)
    ax.set_ylim([0, 100])
    ax.set_xlim([0, 100])
    if not show:
        ax.set_axis_off()
        ax.patch.set_alpha(0)

    delta = -1
    current = 100
    if orientation == "bottom":
        delta = 1
        current = 0

    fontsize = kwargs.pop("fontsize", 16)
    delta = fontsize * delta

    conditions = data.S.experimental_conditions

    def render_polarization(c):
        pol = c["polarization"]
        if pol in ["lc", "rc"]:
            return "\\textbf{" + pol.upper() + "}"

        symbol_pol = {
            "s": "",
            "p": "",
            "s-p": "",
            "p-s": "",
        }

        prefix = ""
        if pol in ["s-p", "p-s"]:
            prefix = "\\textbf{Linear Dichroism, }"

        symbol = symbol_pol[pol]
        if symbol:
            return prefix + "$" + symbol + "$/\\textbf{" + pol + "}"

        return prefix + "\\textbf{" + pol + "}"

    def render_photon(c):
        return "\\textbf{" + str(c["hv"]) + " eV"

    renderers = {
        "temp": lambda c: "\\textbf{T = " + "{:.3g}".format(c["temp"]) + " K}",
        "photon": render_photon,
        "photon polarization": lambda c: render_photon(c) + ", " + render_polarization(c),
        "polarization": render_polarization,
    }

    for item in desc:
        if isinstance(item, (float, int)):
            current += item + delta
            continue

        item = item.replace("_", " ").lower()

        ax.text(0, current, renderers[item](conditions), fontsize=fontsize, **kwargs)
        current += delta


def annotate_cuts(ax, data, plotted_axes, include_text_labels=False, **kwargs):
    """Annotates a cut location onto a plot.

    Example:
        >>> annotate_cuts(ax, conv, ['kz', 'ky'], hv=80)  # doctest: +SKIP

    Args:
        ax: The axes to plot onto
        data: The original data
        plotted_axes: The dimension names which were plotted
        include_text_labels: Whether to include text labels
        kwargs: Defines the coordinates of the cut location
    """
    converted_coordinates = convert_coordinates_to_kspace_forward(data)
    assert len(plotted_axes) == 2

    for k, v in kwargs.items():
        if not isinstance(v, (tuple, list, np.ndarray)):
            v = [v]

        selected = converted_coordinates.sel(**dict([[k, v]]), method="nearest")

        for coords_dict, obj in selected.G.iterate_axis(k):
            css = [obj[d].values for d in plotted_axes]
            ax.plot(*css, color="red", ls="--", linewidth=1, dashes=(5, 5))

            if include_text_labels:
                idx = np.argmin(css[1])

                ax.text(
                    css[0][idx] + 0.05,
                    css[1][idx],
                    "{} = {} {}".format(name_for_dim(k), coords_dict[k].item(), unit_for_dim(k)),
                    color="red",
                    size="medium",
                )


def annotate_point(ax, location, label, delta=None, **kwargs):
    """Annotates a point or high symmetry location into a plot."""
    label = {
        "G": "$\\Gamma$",
        "X": r"\textbf{X}",
        "Y": r"\textbf{Y}",
        "K": r"\textbf{K}",
        "M": r"\textbf{M}",
    }.get(label, label)

    if delta is None:
        delta = (
            -0.05,
            0.05,
        )

    c = kwargs.pop("color", "red")

    if len(delta) == 2:
        dx, dy = tuple(delta)
        x, y = tuple(location)
        ax.plot([x], [y], "o", c=c)
        ax.text(x + dx, y + dy, label, color=c, **kwargs)
    else:
        dx, dy, dz = tuple(delta)
        x, y, z = tuple(location)
        ax.plot([x], [y], [z], "o", c=c)
        ax.text(x + dx, y + dy, z + dz, label, color=c, **kwargs)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np


def safe_reshape(data: "np.ndarray", shape: tuple[int]):
    """
    Attempt to reshape data to given shape. If the data is shorter than the shape, the
    final dimension will be shorter than requested. This is often useful for reshaping
    data from a scan that was terminated early.

    Args:
        data: The data to reshape
        shape: The new shape

    Returns:
        reshaped: the reshaped data
    """
    new_shape = shape[:-1] + (-1,)
    return data.reshape(new_shape)



"""Implements transform pipelines for pytorch_lightning with basic inverse transform."""
from dataclasses import dataclass, field
from typing import Callable, List, Any

__all__ = ["ComposeBoth", "ReversibleLambda", "Identity"]


class Identity:
    """Represents a reversible identity transform."""

    def encodes(self, x):
        return x

    def __call__(self, x):
        return x

    def decodes(self, x):
        return x

    def __repr__(self):
        return "Identity()"


_identity = Identity()


@dataclass
class ReversibleLambda:
    """A reversible anonymous function, so long as the caller supplies an inverse."""

    encodes: Callable = field(repr=False)
    decodes: Callable = field(default=lambda x: x, repr=False)

    def __call__(self, value):
        """Apply the inner lambda to the data in forward pass."""
        return self.encodes(value)


@dataclass
class ComposeBoth:
    """Like `torchvision.transforms.Compose` but this operates on data & target in each transform."""

    transforms: List[Any]

    def __post_init__(self):
        """Replace missing transforms with identities."""
        safe_transforms = []
        for t in self.transforms:
            if isinstance(t, (tuple, list)):
                xt, yt = t
                t = [xt or _identity, yt or _identity]

            safe_transforms.append(t)

        self.original_transforms = self.transforms
        self.transforms = safe_transforms

    def __call__(self, x, y):
        """If this transform has separate data and target functions, apply separately.

        Otherwise, we apply the single transform to both the data and the target.
        """
        for t in self.transforms:
            if isinstance(t, (list, tuple)):
                xt, yt = t
                x, y = xt(x), yt(y)
            else:
                x, y = t(x, y)

        return x, y

    def decodes_target(self, y):
        """Pull the target back in the transform stack as far as possible.

        This is necessary only for the predicted target because
        otherwise we can *always* push the ground truth target and input
        forward in the transform stack.

        This is imperfect because for some transforms we need X and Y
        in order to process the data.
        """
        for t in self.transforms[::-1]:
            if isinstance(t, (list, tuple)):
                _, yt = t

                y = yt.decodes(y)
            else:
                break

        return y

    def __repr__(self):
        """Show both of the constitutent parts of this transform."""
        return (
            self.__class__.__name__
            + "(\n\t"
            + "\n\t".join([str(t) for t in self.original_transforms])
            + "\n)"
        )



"""Tools related to finding the Fermi momentum in a cut."""
import numpy as np

from arpes.fits import LinearModel
from arpes.typing import DataType

__all__ = ("kfermi_from_mdcs",)


def kfermi_from_mdcs(mdc_results: DataType, param=None):
    """Calculates a Fermi momentum using a series of MDCs and the known Fermi level (eV=0).

    This is especially useful to isolate an area for analysis.

    This method tolerates data that came from a prefixed model fit,
    but will always attempt to look for an attribute containing "center".

    Args:
        mdc_results: A DataArray or Dataset containing :obj:``lmfit.ModelResult``s.
        param: The name of the parameter to use as the Fermi momentum in the fit

    Returns:
        The k_fermi values for the input data.
    """
    real_param_name = None

    param_names = mdc_results.results.F.parameter_names

    if param in param_names and param is not None:
        real_param_name = param
    else:
        best_names = [p for p in param_names if "center" in p]
        if param is not None:
            best_names = [p for p in best_names if param in p]

        assert len(best_names) == 1
        real_param_name = best_names[0]

    def nan_sieve(_, x):
        return not np.isnan(x.item())

    return (
        LinearModel()
        .guess_fit(mdc_results.F.p(real_param_name).G.filter_coord("eV", nan_sieve))
        .eval(x=0)
    )



"""Infrastructure code for Qt application windows."""

import sys

from PyQt5 import QtGui, QtCore, QtWidgets

from arpes.settings import SETTINGS
from arpes.utilities.excepthook import patched_excepthook
from arpes.utilities.qt.ui import KeyBinding, Key

__all__ = ("SimpleWindow",)


class SimpleWindow(QtWidgets.QMainWindow):
    """Provides a relatively simple way of making a windowed application.

    The following utilities are largely managed for you:

    1. Keybindings and chords
    2. Help and info
    3. Application lifecycle
    4. Inter-component messaging
    """

    HELP_DIALOG_CLS = None

    def __init__(self, *args, **kwargs):
        """Configures the window.

        In order to start the window, we

        * patch exception handling to provide a nicer traceback format
        * register key bindings and cursor modes
        * install filters to drop unnecessary Qt events
        """
        super().__init__()
        self.app = None  # this will eventually be a weakref to the application
        self._help_dialog = None

        self._old_excepthook = sys.excepthook
        sys.excepthook = patched_excepthook

        self._cursorModes = self.compile_cursor_modes()
        self._keyBindings = self.compile_key_bindings()

        QtGui.QGuiApplication.installEventFilter(self, self)

    def compile_key_bindings(self):
        """Application generic key bindings.

        Additional keybindings can be added here as requied by the tool.
        """
        return [
            KeyBinding("Close Window", [Key.Key_Escape], self.do_close),
            KeyBinding("Toggle Help", [Key.Key_H], self.toggle_help),
        ]

    def compile_cursor_modes(self):
        """Unused hook for supporting additional cursor modes."""
        return []

    def closeEvent(self, event):
        self.do_close(event)

    def do_close(self, event):
        """Handler for closing accepting an unused event arg."""
        self.close()

    def close(self):
        """If we need to close, give the application a chance to clean up first."""
        sys.excepthook = self._old_excepthook
        self.app().close()
        super().close()

    def eventFilter(self, source, event: QtGui.QKeyEvent):
        """Neglect Qt events which do not relate to key presses for now."""
        if event.type() == QtCore.QEvent.KeyPress:
            self.handleKeyPressEvent(event)

        return super().eventFilter(source, event)

    def handleKeyPressEvent(self, event):
        """Listener for key events supporting single key chords."""
        handled = False
        for binding in self._keyBindings:
            for combination in binding.chord:
                # only detect single keypresses for now
                if combination == event.key():
                    handled = True
                    binding.handler(event)

        if not handled:
            if SETTINGS.get("DEBUG", False):
                print(f"{event.key()} @ {type(self)}:{event}")

    def toggle_help(self, event):
        """Open and close (toggle) the help panel for the application."""
        if self.HELP_DIALOG_CLS is None:
            return

        if self._help_dialog is None:
            self._help_dialog = self.HELP_DIALOG_CLS(shortcuts=self._keyBindings)
            self._help_dialog.show()
            self._help_dialog._main_window = self  # pylint: disable=protected-access
        else:
            self._help_dialog.close()
            self._help_dialog = None

    def window_print(self, *args, **kwargs):
        """Forwards prints to the application instance so they end up in Jupyter."""
        print(*args, **kwargs)



import weakref
from math import floor
from xarray import Dataset
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QGridLayout

from arpes.typing import DataType
from arpes.utilities.qt.data_array_image_view import DataArrayImageView, DataArrayPlot
from arpes.utilities.qt.utils import ReactivePlotRecord, PlotOrientation
from .base import ImageTool


class ChildWindow(QWidget):
    def __init__(self, data: DataType):
        super().__init__()
        self.setLayout(QGridLayout())
        self.spectrum = data.S.spectrum if isinstance(data, Dataset) else data


class ParentTool(ImageTool):
    """
    ImageTool with child windows displaying data with coordinate slices taken from the
    parent tool.
    """

    TITLE = "Parent Tool"

    children: list[tuple[ChildWindow, tuple[str, ...]]]
    child_views: list[ReactivePlotRecord]

    def __init__(self):
        super().__init__()
        self.children = []
        self.child_views = []

    def add_child(self, data: DataType, connected_axes: str | tuple[str, ...]):
        child = ChildWindow(data)
        connected_axes = (
            (connected_axes,) if isinstance(connected_axes, str) else connected_axes
        )
        self.children.append((child, connected_axes))

    def generate_child_marginal(
        self,
        child: ChildWindow,
        connected_axes: tuple[str, ...],
        name: str = None,
    ):
        """
        Generates a marginal plot for this applications's data after selecting along
        `dimensions`. This is used to generate the many different views of a volume in
        the browsable tools.
        """
        layout = child.layout()
        spectrum = child.spectrum

        remaining_dims = [dim for dim in spectrum.dims if dim not in connected_axes]

        if len(remaining_dims) == 1:
            widget = DataArrayPlot(
                name=name,
                root=weakref.ref(child),
                orientation=PlotOrientation.Horizontal,
            )
            self.views[name] = widget
            widget.setMaximumHeight(200)
        else:
            assert len(remaining_dims) == 2
            widget = DataArrayImageView(name=name, root=weakref.ref(child))
            widget.view.setAspectLocked(False)
            self.views[name] = widget

            # TODO: set this with the correct ninety_eight_percentile
            # widget.setHistogramRange(0, self.ninety_eight_percentile)
            widget.setLevels(0.05, 0.95)

        self.child_views.append(
            ReactivePlotRecord(
                dims=connected_axes, view=widget, orientation=PlotOrientation.Horizontal
            )
        )
        layout.addWidget(widget, 0, 0)
        return widget

    def update_reactive_views(
        self,
        coord_slices: dict[str, slice] = None,
        changed_dimensions: set[str] = None,
        force=False,
        keep_levels=True,
    ):
        return super().update_reactive_views(
            coord_slices,
            changed_dimensions,
            force,
            keep_levels,
            self.reactive_views + self.child_views,
        )

    def before_show(self):
        for i, (child, connected_axes) in enumerate(self.children):
            self.generate_child_marginal(
                child,
                connected_axes,
                name=f"child_{i}",
            )

        super().before_show()

    def after_show(self):
        TITLE_BAR_HEIGHT = 32
        TASKBAR_HEIGHT = 50
        screen_size = self.primaryScreen().size()
        screen_width = screen_size.width()
        screen_height = screen_size.height() - TASKBAR_HEIGHT
        parent_proportion = 3 / 5

        self.window.setGeometry(
            0,
            TITLE_BAR_HEIGHT,
            floor(screen_width * parent_proportion),
            screen_height - TITLE_BAR_HEIGHT,
        )

        n_axes = {
            child: (len(child.spectrum.dims) - len(connected_axes))
            for child, connected_axes in self.children
        }
        total_child_axes = sum(n_axes.values())
        assert total_child_axes > 0

        starting_height = TITLE_BAR_HEIGHT
        for child_window, n_child_axes in n_axes.items():
            child_window.show()
            child_height = screen_height // total_child_axes * n_child_axes
            child_window.setGeometry(
                floor(screen_width * parent_proportion),
                starting_height,
                floor(screen_width * (1 - parent_proportion)),
                child_height - TITLE_BAR_HEIGHT,
            )
            starting_height += child_height

        return super().after_show()

    def close(self):
        super().close()
        for child, _ in self.children:
            child.close()



"""Implements support for the Lanzara HHG lab."""

import re
import os
from pathlib import Path

import numpy as np
import pandas as pd

import xarray as xr

from arpes.endstations import HemisphericalEndstation
from ..utilities.pxt import read_single_pxt

__all__ = ("HHGEndstation",)


def get_associated_kaindl_files(reference_path: Path):
    name_match = re.match(
        r"([\w+]*_?scan_[0-9][0-9][0-9]_)[0-9][0-9][0-9]\.pxt", reference_path.name
    )

    if name_match is None:
        return [reference_path]

    fragment = name_match.groups()[0]
    components = list(reference_path.parent.glob(f"{fragment}*.pxt"))
    components.sort()

    return components


def read_ai_file(path: Path) -> pd.DataFrame:
    """Read metadata from the Kaindl _AI.txt files.

    Kayla and Conrad discovered that Scienta does not record these files in a
    standardized format, but instead puts an arbitrarily long header at the top of the
    file and sometimes omits the column names.

    By manual inspection, we determined that despite this, the columns appear consistent
    across files recorded in these two formats. The columns are:

    ["Elapsed Time (s)", "Main Chamber", "Garage", "Integrated Photo AI",
     "Photo AI", "Photocurrent", "Heater Power", "Temperature A",
     "Temperature B"]

    depending on whether the header is there or not we need to skip a variable number of
    lines. The way we are detecting this is to look for the presence of the header and
    if it is in the file use it as the previous line before the start of the data.
    Ultimately we defer loading to pandas.

    Otherwise, if the header is absent we look for a tab as the first line of data.
    """
    with open(path, "r") as f:
        lines = f.readlines()

    first_line_no = None
    for i, line in enumerate(lines):
        if "\t" in line:
            first_line_no = i
            break

    column_names = [
        "Elapsed Time (s)",
        "Main Chamber",
        "Garage",
        "Integrated Photo AI",
        "Photo AI",
        "Photocurrent",
        "Heater Power",
        "Temperature A",
        "Temperature B",
    ]

    return pd.read_csv(path, sep="\t", skiprows=first_line_no, names=column_names)


class HHGEndstation(HemisphericalEndstation):
    """The Kaindl Tr-ARPES high harmonic generation setup."""

    PRINCIPAL_NAME = "HHG"
    ALIASES = []

    _TOLERATED_EXTENSIONS = {
        ".pxt",
    }
    _SEARCH_PATTERNS = (
        r"[\-a-zA-Z0-9_\w+]+scan_[0]*{}_[0-9][0-9][0-9]",
        r"[\-a-zA-Z0-9_\w+]+scan_[0]*{}",
    )

    RENAME_KEYS = {
        "Delay Stage": "delay",
    }

    def resolve_frame_locations(self, scan_desc: dict = None):
        """Find .pxt files associated to a potentially multi-cut scan.

        This is very similar to what happens on BL4 at the ALS. You can look at the code
        for MERLIN to see more about how this works, or in
        `find_kaindl_files_associated`.
        """
        if scan_desc is None:
            raise ValueError(
                "Must pass dictionary as file scan_desc to all endstation loading code."
            )

        original_data_loc = scan_desc.get("path", scan_desc.get("file"))
        p = Path(original_data_loc)
        assert p.exists(), f"File {original_data_loc} does not exist."

        return get_associated_kaindl_files(p)

    def load_single_frame(
        self, frame_path: str = None, scan_desc: dict = None, **kwargs
    ):
        data = read_single_pxt(frame_path)
        return xr.Dataset({"spectrum": data}, attrs=data.attrs)

    def concatenate_frames(
        self,
        frames=list[xr.Dataset],
        scan_desc: dict = None,
    ):
        """Concenate frames from individual .pxt files on the Kaindl setup.

        The unique challenge here is to look for and parse the motor positions file (if
        they exist) and add this as a coordinate. As in Beamline 4 at the ALS, these
        Motor_Pos file gives the scan coordinate which we need to concatenate along.
        """
        if len(frames) < 2:
            return super().concatenate_frames(frames)

        # determine which axis to stitch them together along, and then do this
        original_filename = scan_desc.get("path", scan_desc.get("file"))

        internal_match = re.match(
            r"([a-zA-Z0-9\w+_]+)_[0-9][0-9][0-9]\.pxt", Path(original_filename).name
        )
        if internal_match.groups():
            motors_path = str(
                Path(original_filename).parent
                / f"{internal_match.groups()[0]}_Motor_Pos.txt"
            )
            with open(motors_path, "r") as f:
                motor_names = f.readline().strip().split("\t")
            frame_positions = np.loadtxt(motors_path, delimiter="\t", skiprows=1)

            for frame, frame_position in zip(frames, frame_positions):
                for motor_name, motor_position in zip(
                    motor_names, np.atleast_1d(frame_position)
                ):
                    frame.coords[motor_name] = motor_position

            def combine_datasets(ds_list, shape: list, dim_names: list):
                assert len(ds_list) == np.prod(shape)

                for dim_name, subshape in zip(dim_names, shape):
                    ds_list = [
                        xr.concat(
                            ds_list[i : i + subshape], dim_name, coords="different"
                        )
                        for i in range(0, len(ds_list), subshape)
                    ]
                return ds_list[0]

            shape = [
                len(np.unique(individual_motor_positions))
                for individual_motor_positions in frame_positions.T
            ]
            try:
                return combine_datasets(frames, shape, motor_names)
            except AssertionError:
                n_positions = np.prod(shape)
                while n_positions > len(frames):
                    shape[-1] -= 1
                    n_positions = np.prod(shape)
                return combine_datasets(frames[:n_positions], shape, motor_names)

    def postprocess_scan(self, data: xr.Dataset, scan_desc: dict = None):
        """Peform final data preprocessing for the HHG lab Tr-ARPES setup.

        This is very similar to what happens at BL4/MERLIN because the code was adopted
        from an old version of the DAQ on that beamline.
        """
        original_filename = Path(scan_desc.get("path", scan_desc.get("file")))
        internal_match = re.match(
            r"([a-zA-Z0-9\w+_]+_[0-9][0-9][0-9])\.pxt", original_filename.name
        )
        all_filenames = get_associated_kaindl_files(original_filename)
        all_filenames = [
            os.path.join(f.parent, f"{f.stem}_AI.txt") for f in all_filenames
        ]

        def load_attr_for_frame(filename, attr_name):
            # this is rereading which is not ideal but can be adjusted later
            df = read_ai_file(Path(filename))
            return df[attr_name][1:].to_numpy(dtype=float).mean()

        def attach_attr(data: xr.Dataset, attr_name, as_name):
            attributes = np.array(
                [load_attr_for_frame(f, attr_name) for f in all_filenames]
            )

            if len(attributes) == 1:
                data[as_name] = attributes[0]
            else:
                non_spectrometer_dims = [
                    d for d in data.data_vars["spectrum"].dims if d not in {"eV", "phi"}
                ]
                non_spectrometer_coords = {
                    c: v
                    for c, v in data.data_vars["spectrum"].coords.items()
                    if c in non_spectrometer_dims
                }

                new_shape = [len(data.coords[d]) for d in non_spectrometer_dims]
                n_positions = np.prod(new_shape)
                attributes_arr = xr.DataArray(
                    attributes[:n_positions].reshape(new_shape),
                    coords=non_spectrometer_coords,
                    dims=non_spectrometer_dims,
                )

                data = xr.merge([data, xr.Dataset(dict([[as_name, attributes_arr]]))])

            return data

        try:
            data = attach_attr(data, "Photocurrent", "photocurrent")
            data = attach_attr(data, "Temperature B", "temp")
            data = attach_attr(data, "Temperature A", "cryotip_temp")
        except FileNotFoundError as e:
            print(e)

        if internal_match.groups():
            attrs_path = str(
                Path(original_filename).parent / f"{internal_match.groups()[0]}_AI.txt"
            )

            try:
                extra = pd.read_csv(attrs_path, sep="\t", skiprows=6)
                data = data.assign_attrs(extra=extra.to_json())
            except Exception:
                # WELP we tried
                pass

        deg_to_rad_coords = {"theta", "beta", "phi"}

        for c in deg_to_rad_coords:
            if c in data.dims:
                data.coords[c] = data.coords[c] * np.pi / 180

        deg_to_rad_attrs = {"theta", "beta", "alpha", "chi"}
        for angle_attr in deg_to_rad_attrs:
            if angle_attr in data.attrs:
                data.attrs[angle_attr] = float(data.attrs[angle_attr]) * np.pi / 180

        ls = [data] + data.S.spectra
        for l in ls:
            l.coords["x"] = np.nan
            l.coords["y"] = np.nan
            l.coords["z"] = np.nan

        data = super().postprocess_scan(data, scan_desc)

        return data



"""Contains utilities for training machine learning and deep learning models.

This includes prebuilt models for common ARPES tasks, and tools which facilitate integration
with other libraries including pytorch.
"""



"""Utilities related to curve-fitting of ARPES data and xarray format data."""
from .fit_models import *
from .utilities import *

# evaluates our monkeypatching code
from .lmfit_html_repr import *
from .lmfit_plot import *



"""Common implementations of peaks, backgrounds for other models."""
import numpy as np
from scipy.special import erfc  # pylint: disable=no-name-in-module

__all__ = [
    "gaussian",
    "affine_bkg",
    "lorentzian",
    "twolorentzian",
    "fermi_dirac",
    "fermi_dirac_affine",
    "gstepb",
    "gstep",
    "gstep_stdev",
    "band_edge_bkg",
    "g",
]


def affine_bkg(x: np.ndarray, lin_bkg=0, const_bkg=0) -> np.ndarray:
    """An affine/linear background.

    Args:
        x:
        lin_bkg:
        const_bkg:

    Returns:
        Background of the form
          lin_bkg * x + const_bkg
    """
    return lin_bkg * x + const_bkg


def gaussian(x, center=0, sigma=1, amplitude=1):
    """Some constants are absorbed here into the amplitude factor."""
    return amplitude * np.exp(-((x - center) ** 2) / (2 * sigma ** 2))


def g(x, mu=0, sigma=0.1):
    """TODO, unify this with the standard Gaussian definition because it's gross."""
    return (1 / np.sqrt(2 * np.pi * sigma ** 2)) * np.exp(-(1 / 2) * ((x - mu) / sigma) ** 2)


def lorentzian(x, gamma, center, amplitude):
    """A straightforward Lorentzian."""
    return amplitude * (1 / (2 * np.pi)) * gamma / ((x - center) ** 2 + (0.5 * gamma) ** 2)


def fermi_dirac(x, center=0, width=0.05, scale=1):
    """Fermi edge, with somewhat arbitrary normalization."""
    return scale / (np.exp((x - center) / width) + 1)


def gstepb(x, center=0, width=1, erf_amp=1, lin_bkg=0, const_bkg=0):
    """Fermi function convoled with a Gaussian together with affine background.

    This accurately represents low temperature steps where thermal broadening is
    less substantial than instrumental resolution.

    Args:
        x: value to evaluate function at
        center: center of the step
        width: width of the step
        erf_amp: height of the step
        lin_bkg: linear background slope
        const_bkg: constant background

    Returns:
        The step edge.
    """
    dx = x - center
    return const_bkg + lin_bkg * np.min(dx, 0) + gstep(x, center, width, erf_amp)


def gstep(x, center=0, width=1, erf_amp=1):
    """Fermi function convolved with a Gaussian.

    Args:
        x: value to evaluate fit at
        center: center of the step
        width: width of the step
        erf_amp: height of the step

    Returns:
        The step edge.
    """
    dx = x - center
    return erf_amp * 0.5 * erfc(1.66511 * dx / width)


def band_edge_bkg(
    x, center=0, width=0.05, amplitude=1, gamma=0.1, lor_center=0, offset=0, lin_bkg=0, const_bkg=0
):
    """Lorentzian plus affine background multiplied into fermi edge with overall offset."""
    return (lorentzian(x, gamma, lor_center, amplitude) + lin_bkg * x + const_bkg) * fermi_dirac(
        x, center, width
    ) + offset


def fermi_dirac_affine(x, center=0, width=0.05, lin_bkg=0, const_bkg=0, scale=1):
    """Fermi step edge with a linear background above the Fermi level."""
    # Fermi edge with an affine background multiplied in
    return (scale + lin_bkg * x) / (np.exp((x - center) / width) + 1) + const_bkg


def gstep_stdev(x, center=0, sigma=1, erf_amp=1):
    """Fermi function convolved with a Gaussian.

    Args:
        x: value to evaluate fit at
        center: center of the step
        sigma: width of the step
        erf_amp: height of the step
    """
    dx = x - center
    return erf_amp * 0.5 * erfc(np.sqrt(2) * dx / sigma)


def twolorentzian(x, gamma, t_gamma, center, t_center, amp, t_amp, lin_bkg, const_bkg):
    """A double lorentzian model.

    This is typically not necessary, as you can use the
    + operator on the Model instances. For instance `LorentzianModel() + LorentzianModel(prefix='b')`.

    This mostly exists for people that prefer to do things the "Igor Way".

    Args:
        x
        gamma
        t_gamma
        center
        t_center
        amp
        t_amp
        lin_bkg
        const_bkg

    Returns:
        A two peak structure.
    """
    L1 = lorentzian(x, gamma, center, amp)
    L2 = lorentzian(x, t_gamma, t_center, t_amp)
    AB = affine_bkg(x, lin_bkg, const_bkg)
    return L1 + L2 + AB



"""Utilities for gap fitting in ARPES, contains tools to normalize by Fermi-Dirac occupation."""
import warnings

import numpy as np

import xarray as xr
from arpes.constants import K_BOLTZMANN_EV_KELVIN
from arpes.fits.fit_models import AffineBroadenedFD
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ("normalize_by_fermi_dirac", "determine_broadened_fermi_distribution", "symmetrize")


def determine_broadened_fermi_distribution(reference_data: DataType, fixed_temperature=True):
    """Determine the parameters for broadening by temperature and instrumental resolution.

    As a general rule, we first try to estimate the instrumental broadening and linewidth broadening
    according to calibrations provided for the beamline + instrument, as a starting point.

    We also calculate the thermal broadening to expect, and fit an edge location. Then we use a Gaussian
    convolved Fermi-Dirac distribution against an affine density of states near the Fermi level, with a constant
    offset background above the Fermi level as a simple but effective model when away from lineshapes.

    These parameters can be used to bootstrap a fit to actual data or used directly in ``normalize_by_fermi_dirac``.

    Args:
        reference_data: The data we want to estimate from.
        fixed_temperature: Whether we should force the temperature to the recorded value.

    Return:
        The estimated fit result for the Fermi distribution.
    """
    params = {}

    if fixed_temperature:
        params["fd_width"] = {
            "value": reference_data.S.temp * K_BOLTZMANN_EV_KELVIN,
            "vary": False,
        }

    reference_data = normalize_to_spectrum(reference_data)

    sum_dims = list(reference_data.dims)
    sum_dims.remove("eV")

    return AffineBroadenedFD().guess_fit(reference_data.sum(sum_dims), params=params)


@update_provenance("Normalize By Fermi Dirac")
def normalize_by_fermi_dirac(
    data: DataType,
    reference_data: DataType = None,
    plot=False,
    broadening=None,
    temperature_axis=None,
    temp_offset=0,
    **kwargs
):
    """Normalizes data according to a Fermi level reference on separate data or using the same source spectrum.

    To do this, a linear density of states is multiplied against a resolution broadened Fermi-Dirac
    distribution (`arpes.fits.fit_models.AffineBroadenedFD`). We then set the density of states to 1 and
    evaluate this model to obtain a reference that the desired spectrum is normalized by.

    Args:
        data: Data to be normalized.
        reference_data: A reference spectrum, typically a metal
            reference. If not provided the integrated data is used.
            Beware: this is inappropriate if your data is gapped.
        plot: A debug flag, allowing you to view the normalization
            spectrum and relevant curve-fits.
        broadening: Detector broadening.
        temperature_axis: Temperature coordinate, used to adjust the
            quality of the reference for temperature dependent data.
        temp_offset: Temperature calibration in the case of low
            temperature data. Useful if the temperature at the sample is
            known to be hotter than the value recorded off of a diode.
        **kwargs

    Returns:
        Data after normalization by the Fermi occupation factor.
    """
    reference_data = data if reference_data is None else reference_data
    broadening_fit = determine_broadened_fermi_distribution(reference_data, **kwargs)
    broadening = broadening_fit.params["conv_width"].value if broadening is None else broadening

    if plot:
        print(
            "Gaussian broadening is: {} meV (Gaussian sigma)".format(
                broadening_fit.params["conv_width"].value * 1000
            )
        )
        print(
            "Fermi edge location is: {} meV (fit chemical potential)".format(
                broadening_fit.params["fd_center"].value * 1000
            )
        )
        print(
            "Fermi width is: {} meV (fit fermi width)".format(
                broadening_fit.params["fd_width"].value * 1000
            )
        )

        broadening_fit.plot()

    offset = broadening_fit.params["offset"].value
    without_offset = broadening_fit.eval(offset=0)

    cut_index = -np.argmax(without_offset[::-1] > 0.1 * offset)
    cut_energy = reference_data.coords["eV"].values[cut_index]

    if temperature_axis is None and "temp" in data.dims:
        temperature_axis = "temp"

    transpose_order = list(data.dims)
    transpose_order.remove("eV")

    if temperature_axis:
        transpose_order.remove(temperature_axis)
        transpose_order = transpose_order + [temperature_axis]

    transpose_order = transpose_order + ["eV"]

    without_background = (data - data.sel(eV=slice(cut_energy, None)).mean("eV")).transpose(
        *transpose_order
    )

    if temperature_axis:
        without_background = normalize_to_spectrum(without_background)
        divided = without_background.G.map_axes(
            temperature_axis,
            lambda x, coord: x
            / broadening_fit.eval(
                x=x.coords["eV"].values,
                lin_bkg=0,
                const_bkg=1,
                offset=0,
                conv_width=broadening,
                fd_width=(coord[temperature_axis] + temp_offset) * K_BOLTZMANN_EV_KELVIN,
            ),
        )
    else:
        without_background = normalize_to_spectrum(without_background)
        divided = without_background / broadening_fit.eval(
            x=data.coords["eV"].values, conv_width=broadening, lin_bkg=0, const_bkg=1, offset=0
        )

    divided.coords["eV"].values = (
        divided.coords["eV"].values - broadening_fit.params["fd_center"].value
    )
    return divided


def _shift_energy_interpolate(data: DataType, shift=None):
    if shift is not None:
        pass
        # raise NotImplementedError("arbitrary shift not yet implemented")

    data = normalize_to_spectrum(data).S.transpose_to_front("eV")

    new_data = data.copy(deep=True)
    new_axis = new_data.coords["eV"]
    new_values = new_data.values * 0

    if shift is None:
        closest_to_zero = data.coords["eV"].sel(eV=0, method="nearest")
        shift = -closest_to_zero

    stride = data.G.stride("eV", generic_dim_names=False)

    if np.abs(shift) >= stride:
        n_strides = int(shift / stride)
        new_axis = new_axis + n_strides * stride

        shift = shift - stride * n_strides

    new_axis = new_axis + shift

    weight = float(shift / stride)

    new_values = new_values + data.values * (1 - weight)
    if shift > 0:
        new_values[1:] = new_values[1:] + data.values[:-1] * weight
    if shift < 0:
        new_values[:-1] = new_values[:-1] + data.values[1:] * weight

    new_data.coords["eV"] = new_axis
    new_data.values = new_values

    return new_data


@update_provenance("Symmetrize")
def symmetrize(data: DataType, subpixel=False, full_spectrum=False):
    """Symmetrizes data across the chemical potential.

    This provides a crude tool by which
    gap analysis can be performed. In this implementation, subpixel accuracy is achieved by
    interpolating data.

    Args:
        data: Input array.
        subpixel: Enable subpixel correction
        full_spectrum: Returns data above and below the chemical
            potential. By default, only the bound part of the spectrum
            (below the chemical potential) is returned, because the
            other half is identical.

    Returns:
        The symmetrized data.
    """
    data = normalize_to_spectrum(data).S.transpose_to_front("eV")

    if subpixel or full_spectrum:
        data = _shift_energy_interpolate(data)

    above = data.sel(eV=slice(0, None))
    below = data.sel(eV=slice(None, 0)).copy(deep=True)

    l = len(above.coords["eV"])

    zeros = below.values * 0
    zeros[-l:] = above.values[::-1]

    below.values = below.values + zeros

    if full_spectrum:
        if not subpixel:
            warnings.warn("full spectrum symmetrization uses subpixel correction")

        full_data = below.copy(deep=True)

        new_above = full_data.copy(deep=True)[::-1]
        new_above.coords["eV"] = new_above.coords["eV"] * -1

        full_data = xr.concat([full_data, new_above[1:]], dim="eV")

        result = full_data
    else:
        result = below

    return result



"""Contains electron/hole pocket analysis routines."""
import numpy as np
from sklearn.decomposition import PCA

import xarray as xr
from arpes.fits.fit_models import AffineBackgroundModel, LorentzianModel
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.conversion import slice_along_path

from typing import List, Tuple

__all__ = (
    "curves_along_pocket",
    "edcs_along_pocket",
    "radial_edcs_along_pocket",
    "pocket_parameters",
)


def pocket_parameters(data: DataType, kf_method=None, sel=None, method_kwargs=None, **kwargs):
    """Estimates pocket center, anisotropy, principal vectors, and extent in either angle or k-space.

    Since data can be converted forward it is generally advised to do
    this analysis in angle space before conversion if the pocket is not very large.

    Args:
        data: The input kx-ky or 2D angle map.
        kf_method: How to determine k_F for each slice.
        sel: An energy selection window near the Fermi surface.
        method_kwargs: Passed to the kf determination method
        kwargs: Passed to the radial selection method.

    Returns:
        Extracted asymmetry parameters.
    """
    slices, _ = curves_along_pocket(data, **kwargs)  # slices, angles =

    if kf_method is None:
        kf_method = find_kf_by_mdc

    if sel is None:
        sel = {"eV": slice(-0.03, 0.05)}

    kfs = [kf_method(s if sel is None else s.sel(**sel), **(method_kwargs or {})) for s in slices]

    fs_dims = list(data.dims)
    if "eV" in fs_dims:
        fs_dims.remove("eV")

    locations = [
        {d: ss[d].sel(angle=kf, eV=0, method="nearest").item() for d in fs_dims}
        for kf, ss in zip(kfs, slices)
    ]

    location_vectors = [[coord[d] for d in fs_dims] for coord in locations]
    as_ndarray = np.array(location_vectors)

    pca = PCA(n_components=2)
    pca.fit(as_ndarray)

    return {
        "locations": locations,
        "location_vectors": location_vectors,
        "center": {d: np.mean(np.array([coord[d] for coord in locations])) for d in fs_dims},
        "pca": pca.components_,
    }


@update_provenance("Collect EDCs projected at an angle from pocket")
def radial_edcs_along_pocket(
    data: DataType,
    angle,
    inner_radius=0,
    outer_radius=5,
    n_points=None,
    select_radius=None,
    **kwargs
) -> xr.Dataset:
    """Produces EDCs distributed radially along a vector from the pocket center.

    The pocket center should be passed through kwargs via `{dim}={value}`.

    Example:
        I.e. an appropriate call would be

        >>> radial_edcs_along_pocket(spectrum, np.pi / 4, inner_radius=1, outer_radius=4, phi=0.1, beta=0)  # doctest: +SKIP

    Args:
        data: ARPES Spectrum.
        angle: Angle along the FS to cut against.
        inner_radius: The min for the angle/momentum equivalent radial coordinate.
        outer_radius: The max for the angle/momentum equivalent radial coordinate.
        n_points: Number of EDCs, can be automatically inferred.
        select_radius: The radius used for selections along the radial curve.
        kwargs: Used to define the central point.

    Return:
        A 2D array which has an angular coordinate around the pocket center.
    """
    data = normalize_to_spectrum(data)
    fermi_surface_dims = list(data.dims)

    assert "eV" in fermi_surface_dims
    fermi_surface_dims.remove("eV")

    center_point = {k: v for k, v in kwargs.items() if k in data.dims}
    center_as_vector = np.array([center_point.get(d, 0) for d in fermi_surface_dims])

    if n_points is None:
        stride = data.G.stride(generic_dim_names=False)
        granularity = np.mean(np.array([stride[d] for d in fermi_surface_dims]))
        n_points = int(1.0 * (outer_radius - inner_radius) / granularity)

    if n_points <= 0:
        n_points = 10

    primitive = np.array([np.cos(angle), np.sin(angle)])
    far = center_as_vector + outer_radius * primitive
    near = center_as_vector + inner_radius * primitive
    vecs = zip(near, far)

    radius_coord = np.linspace(inner_radius, outer_radius, n_points or 10)

    data_vars = {}
    for d, points in dict(zip(fermi_surface_dims, vecs)).items():
        print(d, points)
        data_vars[d] = xr.DataArray(
            np.array(np.linspace(points[0], points[1], n_points)),
            coords={"r": radius_coord},
            dims=["r"],
        )

    selection_coords = [{k: v[n] for k, v in data_vars.items()} for n in range(n_points)]

    edcs = [
        data.S.select_around(coord, radius=select_radius, fast=True) for coord in selection_coords
    ]

    for r, edc in zip(radius_coord, edcs):
        edc.coords["r"] = r

    data_vars["data"] = xr.concat(edcs, dim="r")

    return xr.Dataset(data_vars, coords=data_vars["data"].coords)


def curves_along_pocket(
    data: DataType, n_points=None, inner_radius=0, outer_radius=5, shape=None, **kwargs
) -> Tuple[List[xr.DataArray], List[float]]:
    """Produces radial slices along a Fermi surface through a pocket.

    Evenly distributes perpendicular cuts along an
    ellipsoid. The major axes of the ellipsoid can be specified by `shape`
    but must be axis aligned.

    The inner and outer radius parameters control the endpoints of the
    resultant slices along the Fermi surface.

    Args:
        data:
        n_points:
        inner_radius:
        outer_radius:
        shape:
        kwargs:

    Returns:
        A tuple of two lists. The first list contains the slices and the second
        the coordinates of each slice around the pocket center.
    """
    data = normalize_to_spectrum(data)

    fermi_surface_dims = list(data.dims)
    if "eV" in fermi_surface_dims:
        fermi_surface_dims.remove("eV")

    center_point = {k: v for k, v in kwargs.items() if k in data.dims}

    center_as_vector = np.array([center_point.get(d, 0) for d in fermi_surface_dims])

    if n_points is None:
        # determine N approximately by the granularity
        n_points = np.min(list(len(data.coords[d].values) for d in fermi_surface_dims))

    stride = data.G.stride(generic_dim_names=False)
    resolution = np.min([v for s, v in stride.items() if s in fermi_surface_dims])

    angles = np.linspace(0, 2 * np.pi, n_points, endpoint=False)

    def slice_at_angle(theta):
        primitive = np.array([np.cos(theta), np.sin(theta)])
        far = center_as_vector + outer_radius * primitive

        return slice_along_path(
            data,
            [dict(zip(fermi_surface_dims, point)) for point in [center_as_vector, far]],
            resolution=resolution,
        )

    slices = [slice_at_angle(theta) for theta in angles]

    max_ang = slices[0].coords["angle"].max().item()

    slices = [
        s.sel(angle=slice(max_ang * (1.0 * inner_radius / outer_radius), None)).isel(
            angle=slice(None, -1)
        )
        for s in slices
    ]

    for ang, s in zip(angles, slices):
        s.coords["theta"] = ang

    # we do not xr.concat because the interpolated angular dim can actually be different on each due
    # to floating point nonsense
    return slices, angles


def find_kf_by_mdc(slice: DataType, offset=0, **kwargs) -> float:
    """Finds the Fermi momentum by curve fitting an MDC.

    Offset is used to control the radial offset from the pocket for studies where
    you want to go slightly off the Fermi momentum.

    Args:
        slice: Input fit data.
        offset: Offset to add to the result
        kwargs: Passed as parameters to the fit routine.

    Returns:
        The fitting Fermi momentum.
    """
    if isinstance(slice, xr.Dataset):
        slice = slice.data

    assert isinstance(slice, xr.DataArray)

    if "eV" in slice.dims:
        slice = slice.sum("eV")

    lor = LorentzianModel()
    bkg = AffineBackgroundModel(prefix="b_")

    result = (lor + bkg).guess_fit(data=slice, params=kwargs)
    return result.params["center"].value + offset


@update_provenance("Collect EDCs around pocket edge")
def edcs_along_pocket(
    data: DataType, kf_method=None, select_radius=None, sel=None, method_kwargs=None, **kwargs
):
    """Collects EDCs around a pocket.

    This consists first in identifying the momenta
    around the pocket, and then integrating small windows around each of these points.

    Args:
        data: The input kx-ky or 2D angle map.
        kf_method: How to determine k_F for each slice.
        select_radius: The radius used for selections along the radial curve.
        sel: An energy selection window near the Fermi surface.
        method_kwargs: Passed to the kf determination method
        kwargs: Passed to the radial selection method.

    Returns:
        EDCs at the fermi momentum around a pocket.
    """
    slices, angles = curves_along_pocket(data, **kwargs)

    if kf_method is None:
        kf_method = find_kf_by_mdc

    if sel is None:
        sel = {"eV": slice(-0.05, 0.05)}

    kfs = [kf_method(s if sel is None else s.sel(**sel), **(method_kwargs or {})) for s in slices]

    fs_dims = list(data.dims)
    if "eV" in fs_dims:
        fs_dims.remove("eV")

    locations = [
        {d: ss[d].sel(angle=kf, eV=0, method="nearest").item() for d in fs_dims}
        for kf, ss in zip(kfs, slices)
    ]

    edcs = [data.S.select_around(l, radius=select_radius, fast=True) for l in locations]

    data_vars = {}
    index = np.array(angles)

    for d in fs_dims:
        data_vars[d] = xr.DataArray(
            np.array([l[d] for l in locations]), coords={"theta": index}, dims=["theta"]
        )

    for ang, edc in zip(angles, edcs):
        edc.coords["theta"] = ang

    data_vars["spectrum"] = xr.concat(edcs, dim="theta")

    return xr.Dataset(data_vars, coords=data_vars["spectrum"].coords)



"""A live momentun conversion tool, useful for finding and setting offsets."""

from PyQt5 import QtWidgets
import numpy as np

from arpes.utilities import normalize_to_spectrum, group_by
from arpes.typing import DataType
from arpes.utilities.conversion import convert_to_kspace
from arpes.utilities.qt import qt_info, SimpleApp, SimpleWindow
from arpes.utilities.qt.ui import (
    tabs,
    horizontal,
    vertical,
    label,
    numeric_input,
    CollectUI,
)
from arpes.plotting.bz import segments_standard


__all__ = (
    "KTool",
    "ktool",
)


qt_info.setup_pyqtgraph()


class KTool(SimpleApp):
    """Provides a live momentum converting tool.

    QtTool is an implementation of Image/Bokeh Tool based on PyQtGraph and PyQt5 for now we retain
    a number of the metaphors from BokehTool, including a "context" that stores the state, and can
    be used to programmatically interface with the tool.
    """

    TITLE = "KSpace-Tool"
    WINDOW_SIZE = (5, 6)
    WINDOW_CLS = SimpleWindow

    DEFAULT_COLORMAP = "viridis"

    def __init__(
        self, apply_offsets=True, zone=None, energy_window=slice(-0.05, 0.05), **kwargs
    ):
        """Set attributes to safe defaults and unwrap the Brillouin zone definition."""
        super().__init__()

        if isinstance(
            zone,
            (tuple, list),
        ):
            self.segments_x, self.segments_y = zone
        elif zone:
            self.segments_x, self.segments_y = segments_standard(zone)
        else:
            self.segments_x, self.segments_y = None, None

        self.energy_window = energy_window
        self.conversion_kwargs = kwargs
        # self.data = None
        self.content_layout = None
        self.main_layout = None
        self.apply_offsets = apply_offsets

    def configure_image_widgets(self):
        """We have two marginals because we deal with Fermi surfaces, they get configured here."""
        self.generate_marginal_for(
            (), 0, 0, "xy", cursors=False, layout=self.content_layout
        )
        self.generate_marginal_for(
            (), 1, 0, "kxy", cursors=False, layout=self.content_layout
        )

    def add_contextual_widgets(self):
        """The main UI layout for controls and tools."""
        convert_dims = ["theta", "beta", "phi", "psi"]
        if "eV" not in self.data.dims:
            convert_dims += ["chi"]
        if "hv" in self.data.dims:
            convert_dims += ["hv"]

        ui = {}
        with CollectUI(ui):
            controls = tabs(
                [
                    "Controls",
                    horizontal(
                        *[
                            vertical(
                                *[
                                    vertical(
                                        label(p),
                                        numeric_input(
                                            self.data.attrs.get(f"{p}_offset", 0.0),
                                            input_type=float,
                                            id=f"control-{p}",
                                        ),
                                    )
                                    for p in pair
                                ]
                            )
                            for pair in group_by(2, convert_dims)
                        ]
                    ),
                ]
            )

        def update_dimension_name(dim_name):
            def updater(value):
                self.update_offsets(dict([[dim_name, float(value)]]))

            return updater

        for dim in convert_dims:
            ui[f"control-{dim}"].subject.subscribe(update_dimension_name(dim))

        controls.setFixedHeight(qt_info.inches_to_px(1.75))

        self.main_layout.addLayout(self.content_layout, 0, 0)
        self.main_layout.addWidget(controls, 1, 0)

    def update_offsets(self, offsets):
        """Pushes offsets to the display data and optionally, the original data."""
        self.data.S.apply_offsets(offsets)
        if self.apply_offsets:
            self.original_data.S.apply_offsets(offsets)
        self.update_data()

    def layout(self):
        """Initialize the layout components."""
        self.main_layout = QtWidgets.QGridLayout()
        self.content_layout = QtWidgets.QGridLayout()
        return self.main_layout

    def update_data(self):
        """The main redraw method for this tool.

        Converts data into momentum space and populates both the angle-space and momentum
        space views.

        If a Brillouin zone was requestd, plots that over the data as well.
        """
        self.views["xy"].setImage(self.spectrum)

        kdata = convert_to_kspace(self.data, **self.conversion_kwargs)
        if "eV" in kdata.dims:
            kdata = kdata.S.transpose_to_back("eV")

        self.views["kxy"].setImage(kdata.S.nan_to_num())
        if self.segments_x is not None:
            bz_plot = self.views["kxy"].plot_item
            kx, ky = self.conversion_kwargs["kx"], self.conversion_kwargs["ky"]
            for segx, segy in zip(self.segments_x, self.segments_y):
                bz_plot.plot(
                    (segx - kx[0]) / (kx[1] - kx[0]), (segy - ky[0]) / (ky[1] - ky[0])
                )

    def before_show(self):
        """Lifecycle hook for configuration before app show."""
        self.configure_image_widgets()
        self.add_contextual_widgets()
        if self.DEFAULT_COLORMAP is not None:
            self.set_colormap(self.DEFAULT_COLORMAP)

    def after_show(self):
        """Initialize application state after app show. Just redraw."""
        self.update_data()

    def set_data(self, data: DataType):
        """Sets the current data to a new value and resets binning.

        Above what happens in QtTool, we try to extract a Fermi surface, and
        repopulate the conversion.
        """
        original_data = normalize_to_spectrum(data)
        self.original_data = original_data

        if len(data.dims) > 2:
            assert "eV" in original_data.dims
            data = data.sel(eV=self.energy_window).sum("eV", keep_attrs=True)
            data.coords["eV"] = 0
        else:
            data = original_data

        if "eV" in data.dims:
            data = data.S.transpose_to_back("eV")

        self.data = data.copy(deep=True)

        if not self.conversion_kwargs:
            rng_mul = 1
            if data.coords["hv"] < 12:
                rng_mul = 0.5
            if data.coords["hv"] < 7:
                rng_mul = 0.25

            if "eV" in self.data.dims:
                self.conversion_kwargs = {
                    "kp": np.linspace(-2, 2, 400) * rng_mul,
                }
            else:
                self.conversion_kwargs = {
                    "kx": np.linspace(-2, 2, 300) * rng_mul,
                    "ky": np.linspace(-2, 2, 300) * rng_mul,
                }


def ktool(data: DataType, **kwargs):
    """Start the momentum conversion tool."""
    tool = KTool(**kwargs)
    tool.set_data(data)
    tool.start()
    return tool



"""Worker for process parallel curve fitting with `lmfit`.

Uses dill for IPC due to issues with pickling `lmfit` instances.
"""
import xarray as xr
from typing import Any, List, Optional
from .broadcast_common import apply_window, compile_model, unwrap_params
from dataclasses import dataclass, field
import dill

__all__ = ["MPWorker"]


@dataclass
class MPWorker:
    """Worker for performing curve fitting on a high dimensional dataset.

    Essentially represents a closure over the curve fitting parameters (the input data,
    model specification, etc.) which do not change for a batch of curve fits.

    This worker can then be treated as a function of coordinates only which:
    1. Subselects its data at those coordinates
    2. Performs the fit at those coordinates using the retained settings
    3. Returns the fit result

    There are also a few details related to serialization because we are dealing with
    IPC whenever we use multiprocessing.
    """

    data: xr.DataArray
    uncompiled_model: Any

    prefixes: Optional[List[str]]
    params: Any

    safe: bool = False
    serialize: bool = False
    weights: Optional[xr.DataArray] = None
    window: Optional[xr.DataArray] = None

    _model: Any = field(init=False)

    def __post_init__(self):
        """Indicate that the model has not been compiled yet."""
        self._model = None

    @property
    def model(self):
        """Compiles and caches the model used for curve fitting.

        Because of pickling constraints, we send model specifications
        not model instances out to workers and let them compile the model.

        This also ends up being slightly more efficient because the specification
        for a model is just references to classes in code.
        """
        if self._model is not None:
            return self._model

        self._model = compile_model(
            self.uncompiled_model, params=self.params, prefixes=self.prefixes
        )
        self._model.make_params()

        return self._model

    @property
    def fit_params(self):
        """Builds or fetches the parameter hints from closed over attributes."""
        if isinstance(self.params, (list, tuple)):
            return {}

        return self.params

    def __call__(self, cut_coords):
        """Performs a curve fit at the coordinates specified by `cut_coords`."""
        current_params = unwrap_params(self.fit_params, cut_coords)
        cut_data, original_cut_data = apply_window(self.data, cut_coords, self.window)

        if self.safe:
            cut_data = cut_data.G.drop_nan()

        weights_for = None
        if self.weights is not None:
            weights_for = self.weights.sel(**cut_coords)

        try:
            fit_result = self.model.guess_fit(cut_data, params=current_params, weights=weights_for)
        except ValueError:
            fit_result = None

        if fit_result is None:
            true_residual = None
        elif self.window is None:
            true_residual = fit_result.residual
        else:
            true_residual = original_cut_data - fit_result.eval(
                x=original_cut_data.coords[original_cut_data.dims[0]].values
            )

        if self.serialize:
            fit_result = dill.dumps(fit_result)

        return fit_result, true_residual, cut_coords



"""A utility excepthook for Qt applications which ensures errors are visible in Jupyter."""
import traceback
import sys
from collections import namedtuple

__all__ = ("patched_excepthook",)


def patched_excepthook(exc_type, exc_value, exc_tb):
    """Prints the traceback instead of dying silently. Useful for debugging Qt apps in Jupyter."""
    enriched_tb = _add_missing_frames(exc_tb) if exc_tb else exc_tb
    traceback.print_exception(exc_type, exc_value, enriched_tb)


def _add_missing_frames(tb):
    result = fake_tb(tb.tb_frame, tb.tb_lasti, tb.tb_lineno, tb.tb_next)
    frame = tb.tb_frame.f_back
    while frame:
        result = fake_tb(frame, frame.f_lasti, frame.f_lineno, result)
        frame = frame.f_back
    return result


fake_tb = namedtuple("fake_tb", ("tb_frame", "tb_lasti", "tb_lineno", "tb_next"))



"""Some very simple string manipulation utilities."""
import re

__all__ = [
    "snake_case",
    "safe_decode",
]


def snake_case(input: str):
    """Approximately converts a string to python_case."""
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", input)
    s2 = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
    return (
        s2.replace("(", "").replace(")", "").replace(" ", "_").replace("/", "_").replace("__", "_")
    )


def safe_decode(input: bytes, prefer: str = None) -> str:
    """Tries different byte interpretations for decoding... very lazy."""
    codecs = ["utf-8", "latin-1", "ascii"]

    if prefer:
        codecs = [prefer] + codecs

    for codec in codecs:
        try:
            return input.decode(codec)
        except UnicodeDecodeError:
            pass

    input.decode("utf-8")  # COULD NOT DETERMINE CODEC, RAISE



"""Provides a convenience data loader for example data.

Providing example data is essential for ensuring approachability,
but in the past we only provided a single ARPES cut. We now provide
a variety but need to be parsimonious about disk space for downloads.
As a result, this custom loader let's us pretend we store the data in
a higher quality format.
"""

from os import replace
import numpy as np
import xarray as xr
from arpes.endstations import SingleFileEndstation, HemisphericalEndstation
import arpes.xarray_extensions

__all__ = ["ExampleDataEndstation"]


class ExampleDataEndstation(SingleFileEndstation, HemisphericalEndstation):
    """Loads data from exported .nc format saved by xarray. Used for storing example data."""

    PRINCIPAL_NAME = "example_data"

    _TOLERATED_EXTENSIONS = {".nc"}

    def load_single_frame(
        self, frame_path: str = None, scan_desc: dict = None, **kwargs
    ) -> xr.Dataset:
        """Loads single file examples.

        Additionally, copies coordinate offsets onto the dataset because we have
        preloaded these for convenience on maps.
        """
        data = xr.open_dataarray(frame_path)
        data = data.astype(np.float64)

        # Process coordinates so that there are no non-dimension coordinates
        # which are not a function of some index. This is for simplicity for beginners.
        replacement_coords = {}
        for cname, coord in data.coords.items():
            if len(coord.values.shape) and cname not in data.dims:
                replacement_coords[cname] = coord.mean().item()

        data = data.assign_coords(**replacement_coords)

        # Wrap into a dataset
        dataset = xr.Dataset({"spectrum": data})
        dataset.S.apply_offsets(data.S.offsets)

        return dataset



r"""Contains models and utilities for "curve fitting" the forward momentum conversion function.

This allows for rapid fine tuning of the angle offsets used in momentum conversion.

If the user calculates, places, or curve fits for the high symmetry
locations in the data $H_i(\phi,\psi,\theta,\beta,\chi)$, these can be used
as waypoints to find a set of $\Delta\phi$, $\Delta\theta$, $\Delta\chi$, etc.
that minimize

$$
\sum_i \text{min}_j |text{P}(H_i, \Delta\phi, \Delta\theta, \ldots) - S_j|^2
$$

where $S_j$ enumerates the high symmetry points of the known Brillouin zone, and $\text{P}$
is the function that maps forwards from angle to momenta. This can also
be used to calculate moiré information, but using the (many) available
high symmetry points of the moiré superlattice to get a finer estimate of
relative angle alignment, lattice incommensuration, and strain than is possible
using the constituent lattices and Brillouin zones.
"""

import numpy as np

from arpes.utilities.conversion.forward import convert_coordinates


__all__ = ("HighSymmetryPointModel",)


def k_points_residual(paramters, coords_dataset, high_symmetry_points, dimensionality=2):
    momentum_coordinates = convert_coordinates(coords_dataset)
    if dimensionality == 2:
        return np.asarray(
            [
                np.diagonal(momentum_coordinates.kx.values),
                np.diagonal(momentum_coordinates.ky.values),
            ]
        )
    else:
        return np.asarray(
            [
                np.diagonal(momentum_coordinates.kx.values),
                np.diagonal(momentum_coordinates.ky.values),
                np.diagonal(momentum_coordinates.kz.values),
            ]
        )


def minimum_forward_error(
    coordinate_samples,
    phi_offset=0,
    psi_offset=0,
    theta_offset=0,
    beta_offset=0,
    chi_offset=0,
    high_symmetry_points=None,
):
    r"""Sets offsets for a coordinate sample dataset before converting them all to momentum.

    Then, for each sample, the closest high symmetry point among the provided
    `high_symmetry_points` is calculated, and the distance to the high symmetry point obtained.
    The distance of each of the coordinate samples to these symmetry points is returned, and the
    optimizer adjusts the offsets to find a "best" set in the sense of least total L2 distance
    to the symmetry points.

    If the coordinate samples are labelled as described above as H_i, then we return

    $$
    \text{min}_j |text{P}(H_i, \Delta\phi, \Delta\theta, \ldots) - S_j|
    $$

    and the optimizer attempts to optimize for

    $$
    \sum_i \left((\text{min}_j |text{P}(H_i, \Delta\phi, \Delta\theta, \ldots) - S_j|^2)\right)^2.
    $$

    We can therefore control the metric by returning a different distance back to the optimizer, if desired.
    For instance, the L1 distance can be optimized if desired by instead returning

    $$
    \text{min}_j |text{P}(H_i, \Delta\phi, \Delta\theta, \ldots) - S_j|^\frac{1}{2}
    $$

    Args:
        coordinate_samples: (N, 6 + 1)
        phi_offset
        psi_offset
        theta_offset
        beta_offset
        chi_offset
        high_symmetry_points

    Returns:
        The MSE distance error for this set of coordinate offsets.
    """
    pass



"""Preliminary detector window corrections."""
import scipy.interpolate
import xarray as xr
import numpy as np
from typing import List, Dict
import itertools

__all__ = ("DetectorCalibration",)


def build_edge_from_list(points: List[Dict[str, float]]):
    """Converts from a list of edge waypoints to a common representation as a DataSet."""
    dimensions = set(itertools.chain(*[p.keys() for p in points]))
    arrays = {}

    for dim in dimensions:
        values = [p[dim] for p in points]
        arrays[dim] = xr.DataArray(values, coords=dict([[dim, values]]), dims=[dim])
    return xr.Dataset(arrays)


class DetectorCalibration:
    """A detector calibration model allowing for correcting the trapezoidal windowing."""

    _left_edge: xr.Dataset
    _right_edge: xr.Dataset

    def __init__(self, left, right):
        """Build the edges for the calibration from a path for the left and right sides."""
        assert set(left[0].keys()) == {
            "phi",
            "eV",
        }
        self._left_edge = build_edge_from_list(left)
        self._right_edge = build_edge_from_list(right)

        # if the edges were passed incorrectly then do it ourselves
        if self._left_edge.phi.mean() > self._right_edge.phi.mean():
            self._left_edge, self._right_edge = self._right_edge, self._left_edge

    def __repr__(self):
        """Representation showing detailed attributes on edge locations."""
        rep = f"<DetectorCalibration>\n\n"
        rep += "Left Edge\n"
        rep += str(self._left_edge)
        rep += "\n\nRightEdge\n"
        rep += str(self._right_edge)
        return rep

    def correct_detector_angle(self, eV, phi):
        """Applies a calibration to the detector `phi` angle."""
        left, right = (
            scipy.interpolate.interp1d(self._left_edge.eV.values, self._left_edge.phi.values)(
                0
            ).item(),
            scipy.interpolate.interp1d(self._right_edge.eV.values, self._right_edge.phi.values)(
                0
            ).item(),
        )
        xs = np.concatenate([self._left_edge.eV.values, self._right_edge.eV.values])
        ys = np.concatenate([self._left_edge.phi.values, self._right_edge.phi.values])
        zs = np.concatenate(
            [self._left_edge.eV.values * 0 + left, self._right_edge.eV.values * 0 + right]
        )

        return scipy.interpolate.griddata(np.stack([xs, zs], axis=1), ys, (eV, phi))



"""Implements dynamic plugin selection when users do not specify the location for their data."""

import warnings

from arpes.endstations import EndstationBase, resolve_endstation

__all__ = ("FallbackEndstation",)

AUTOLOAD_WARNING = (
    "PyARPES has chosen {} for your data since `location` was not specified. "
    "If this is not correct, ensure the `location` key is specified. Read the plugin documentation "
    "for more details."
)


class FallbackEndstation(EndstationBase):
    """Sequentially tries different loading plugins.

    Different from the rest of the data loaders. This one is used when there is no location specified
    and attempts sequentially to call a variety of standard plugins until one is found that works.
    """

    PRINCIPAL_NAME = "fallback"
    ALIASES = []

    ATTEMPT_ORDER = [
        "ALS-BL7",
    ]

    @classmethod
    def determine_associated_loader(cls, file, scan_desc):
        """Determines which loading plugin to use for a given piece of data.

        This is done by looping through loaders in a predetermined priority order,
        and asking each whether it is capable of loading the file.
        """

        for location in cls.ATTEMPT_ORDER:
            try:
                endstation_cls = resolve_endstation(False, location=location)
                if endstation_cls.is_file_accepted(file, scan_desc):
                    return endstation_cls
            except:
                pass

        raise ValueError(
            f"PyARPES failed to find a plugin acceptable for {file}, \n\n{scan_desc}."
        )

    def load(self, scan_desc: dict = None, file=None, **kwargs):
        """Delegates to a dynamically chosen plugin for loading."""
        if file is None:
            file = scan_desc["file"]

        associated_loader = FallbackEndstation.determine_associated_loader(
            file, scan_desc
        )

        try:
            file = int(file)
            file = associated_loader.find_first_file(file, scan_desc)
            scan_desc["file"] = file
        except ValueError:
            pass

        warnings.warn(AUTOLOAD_WARNING.format(associated_loader))
        return associated_loader().load(scan_desc, **kwargs)

    @classmethod
    def find_first_file(cls, file, scan_desc, allow_soft_match=False):
        """Finds any file associated to this scan.

        Instead actually using the superclass code here, we first try to determine
        which loading pluging should be used. Then, we delegate to that class to
        find the first associated file.
        """
        associated_loader = cls.determine_associated_loader(file, scan_desc)
        warnings.warn(AUTOLOAD_WARNING.format(associated_loader))
        return associated_loader.find_first_file(file, scan_desc, allow_soft_match)



"""Reference plots, for preliminary analysis."""
import warnings

import pandas as pd

import xarray as xr
from arpes.io import load_data
from arpes.preparation import normalize_dim
from arpes.utilities.conversion import convert_to_kspace

__all__ = ["make_reference_plots"]


def make_reference_plots(df: pd.DataFrame = None, with_kspace=False):
    """Makes standard reference plots for orienting oneself."""
    try:
        df = df[df.spectrum_type != "xps_spectrum"]
    except TypeError:
        warnings.warn("Unable to filter out XPS files, did you attach spectra type?")

    # Make scans indicating cut locations
    for index, row in df.iterrows():
        try:
            scan = load_data(index)

            if isinstance(scan, xr.Dataset):
                # make plot series normalized by current:
                scan.S.reference_plot(out=True)
            else:
                scan.S.reference_plot(out=True, use_id=False)

                if scan.S.spectrum_type == "spectrum":
                    # Also go and make a normalized version
                    normed = normalize_dim(scan, "phi")
                    normed.S.reference_plot(out=True, use_id=False, pattern="{}_norm_phi.png")

                    if with_kspace:
                        normalized = normalize_dim(scan, "hv")
                        kspace_converted = convert_to_kspace(normalized)
                        kspace_converted.S.reference_plot(
                            out=True, use_id=False, pattern="k_{}.png"
                        )

                        normed_k = normalize_dim(kspace_converted, "kp")
                        normed_k.S.reference_plot(
                            out=True, use_id=False, pattern="k_{}_norm_kp.png"
                        )

        except Exception as e:
            print(str(e))
            warnings.warn("Cannot make plots for {}".format(index))



"""Utilities and an example of how to make an animated plot to export as a movie."""

from pathlib import Path
from warnings import warn
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import PillowWriter, FuncAnimation

import xarray as xr
from arpes.utilities import lift_spectrum

__all__ = ("plot_movie",)


@lift_spectrum
def plot_movie(
    spectrum: xr.DataArray,
    time_dim: str,
    framerate: int = 10,
    ax=None,
    save_path: Path | str | None = None,
    **kwargs,
):
    """Make an animated plot of a 3D dataset using one dimension as `time`.

    Args
    ----
    spectrum : xr.DataArray
        The 3D dataset to animate.
    time_dim : str
        The name of the dimension to animate over.
    framerate : int (default, 10)
        The number of frames per second.
    ax : plt.Axes (default, None)
        The axes to plot on. If None, a new figure is created.
    save_path : Path | str | None (default, None)
        The path to save the movie to. If None, the movie is not saved.

    Returns
    -------
    animation : FuncAnimation
        The animation object. If `save_path` is not None, the movie is saved to that
        path.
    """

    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 7))
    else:
        fig = ax.get_figure()

    vmax = spectrum.max().item()
    vmin = spectrum.min().item()

    cmap = plt.rcParams["image.cmap"]
    if spectrum.S.is_subtracted:
        cmap = "RdBu"
        vmax = np.max([np.abs(vmin), np.abs(vmax)])
        vmin = -vmax
    vmax = kwargs.get("vmax", vmax)
    vmin = kwargs.get("vmin", vmin)

    def animate(i):
        ax.clear()
        plot = spectrum.isel(**{time_dim: i}).plot(ax=ax, add_colorbar=False)
        return plot

    animation = FuncAnimation(
        fig,
        animate,
        frames=spectrum.sizes[time_dim],
        interval=1000 / framerate,
        **kwargs,
    )

    if save_path is not None:
        save_path = Path(save_path) if isinstance(save_path, str) else save_path
        valid_extension = ".gif"
        if save_path.suffix != valid_extension:
            save_path = save_path.with_suffix(valid_extension)
            warn(
                f"save_path should have a {valid_extension} extension. Saving to {save_path} "
                "instead."
            )
        writer = PillowWriter(fps=framerate)
        animation.save(save_path, writer=writer)

    return animation
